'''The PipelineStep object.

These are manipulable Steps in a PipelineTemplate.
'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import threading
from typing import Any, Dict, List, Optional

from typing_extensions import Protocol

from ...algorithms.impl.algorithm import (Algorithm, AlgorithmCatalog,
                                          AlgorithmCatalogStub)
from ...generator.designator import Designator, StepDesignator

# pylint: disable=protected-access


class GeneratorInterface(Protocol):
    '''Interface for a Generator

    We are avoiding circular imports.

    The type should be
      generate(self, pipeline: PipelineTemplate) -> Dict[Designator, BoundPipeline]
    '''
    def generate(self, pipeline: Any) -> Dict[Designator, Any]:
        '''Generates Dict[Designator, BoundPipeline] for a PipelineTemplate'''


class Error(Exception):
    '''Base error for PipelineStep'''


class AlreadyDefinedError(Error):
    '''Attempt to set a value that can only be set once.'''


class BadValueError(Error, ValueError):
    '''A parameter was passed that makes no sense.'''


# pylint: disable=too-many-instance-attributes
class PipelineStep:
    '''A single step in a pipeline template

    Default hyperparams from the model are overridden by the
    PipelineStep constructor (for use in templates). These in
    turn are overridden by values specified in the Problem Definition.

    If hyperparams are passed in using the 'overrides' argument, rather than kwargs,
        they will contribute to the filename of this step when it is saved.
    '''
    _model: Optional[Algorithm]
    _save_output: Optional[str]
    _hyperparams: Dict[str, Any]
    _overrides: Optional[Dict[str, Any]]
    _algorithm_catalog: AlgorithmCatalog
    _generator: Optional[GeneratorInterface]
    _queried: bool
    _name: Optional[str] = None
    _sn_lock = threading.Lock()
    _serial_number: int = 0
    _serialized_model: Optional[bytes] = None

    @classmethod
    def _assign_sn(cls):
        '''Assign a unique serial number to this step.

        This number is used to ensure each step has a unique filename.'''
        with cls._sn_lock:
            retval = cls._serial_number
            cls._serial_number += 1
        return retval

    @classmethod
    def reset_serial_number(cls):
        '''Set serial number back to 0 for testing.

        Do not use this outside of tests.
        '''
        with cls._sn_lock:
            cls._serial_number = 0

    def __init__(self,  # pylint: disable=duplicate-code, too-many-arguments
                 model: Optional[Algorithm] = None,
                 algorithm_catalog: Optional[AlgorithmCatalog] = None,
                 generator: Optional[GeneratorInterface] = None,
                 name: Optional[str] = None,
                 serialized_model: Optional[bytes] = None,
                 overrides: Optional[Dict[str, Any]] = None,
                 **hyperparams):
        self._model = model
        self._overrides = overrides
        self._algorithm_catalog = algorithm_catalog or AlgorithmCatalogStub()
        self._generator = generator
        self._queried = False
        self._name = name or self._name
        self._serialized_model = serialized_model

        self._serial_number = self._assign_sn()

        assert isinstance(model, Algorithm) or model is None
        if model is not None:
            current_overrides = {}
            current_overrides.update(hyperparams)
            if overrides is not None:
                current_overrides.update(overrides)
            self._hyperparams = model.hyperparams(**current_overrides)
        elif len(hyperparams) != 0:
            raise BadValueError(f'hyperparameter overrides specified with no model: {hyperparams}')
        else:
            # Case: parallel or query step
            self._hyperparams = {}

    @property
    def locked(self) -> bool:
        '''Does this step have a model instance that is currently locked?

        For base PipelineStep, this always returns False as there is no model instance.
        '''
        return False

    def converged(self) -> bool:
        '''If this step has a distributed model, has it converged?

        Always returns True for steps with no models or a non-distributed model.
        '''
        return True

    def set_name(self, name: str) -> 'PipelineStep':
        '''Give this step a name.'''
        if self._name is not None:
            raise AlreadyDefinedError(
                f'can not set name to {name.lower()};'
                f' already set to {self._name}')
        self._name = name.lower()
        return self

    def mark_queried(self) -> 'PipelineStep':
        '''Mark this step as the result of a query step.'''
        self._queried = True
        return self

    @property
    def queried(self) -> bool:
        '''Was this step generated from a query step?'''
        return self._queried

    @property
    def pipeline_designator_component(self) -> str:
        '''Our pipeline_designator_component is generally the name of our model.'''
        if self._name is not None:
            return self._name
        if self._model is not None:
            return self._model.name
        return '_'

    @property
    def opt_name(self) -> Optional[str]:
        '''Give raw access to _name.'''
        return self._name

    @property
    def filename(self) -> Optional[StepDesignator]:
        '''Filename to use when saving this step as part of a trained pipeline.

        If this step does not have a name assigned by the generator or searcher,
            its model name and serial number are used.
        '''
        if self._model is None:
            # We don't save parallel steps or query steps
            return None

        overrides = {}
        if self._overrides is not None:
            overrides.update(self._overrides)
        bindings = sorted([
            f'{hyperparam}={str(value)}'
            for hyperparam, value in overrides.items()])

        suffix = f':{",".join(bindings)}' if len(bindings) > 0 else ''
        retval = f'@{self._model.name}_{self._serial_number}@{self._name or ""}{suffix}'

        return StepDesignator(retval.lower())

    @property
    def serial_number(self) -> int:
        '''Serial number used to ensure that this step's designator is unique'''
        return self._serial_number

    def has_model(self) -> bool:
        '''Does the step have an associated model?'''
        return self._model is not None

    @property
    def model_name(self) -> Optional[str]:
        '''If we have a model, what is its name?'''
        # TODO(Merritt): rename to algorithm_name
        if self._model is None:
            return None
        return self._model.name

    @property
    def model(self):
        '''Provide access to our associated model.'''
        assert self._model is not None, (
            f'BUG: attempt to get unbound model from step {self.pipeline_designator_component}'
        )
        return self._model

    @property
    def serialized_model(self) -> Optional[bytes]:
        '''Provide access to any serialized model state.

        This is used during instatiation to preload model state.
        '''
        return self._serialized_model

    @property
    def tags(self) -> Optional[Dict[str, List[str]]]:
        '''Return the tags from our model if we have one.'''
        if self._model is None:
            return None
        return self._model.tags

    def hyperparams(self, **overrides) -> Dict[str, Any]:
        '''Get the hyperparameters for this step.

        Arguments passed to this function override defaults for the template step.
        '''
        default_hyperparams = self._hyperparams.copy()
        default_hyperparams.update(**overrides)
        return default_hyperparams

    def generate(self,
                 future_steps: List['PipelineStep']
                 ) -> Dict[Designator, List['PipelineStep']]:
        '''Generate pipeline sketches (lists of bound steps) for this step and all subsequent steps.

        Each query step may expand into multiple 'pipeline sketches'
        '''
        if len(future_steps) == 0:
            return {Designator(): [self]}
        step0 = future_steps[0]
        sketches = step0.generate(future_steps[1:])
        return {
            Designator(key): [self] + rest[:]
            for key, rest in sketches.items()
        }

    def clone(self, **overrides) -> 'PipelineStep':
        '''Clone this step, providing hyperparameter overrides.

        'overrides' argument will contribute to step filename when it is saved.

        Pre-existing hyperparams on this step will be passed along to
        the child step but will not contribute to its filename.
        '''
        child_overrides = {}
        if self._overrides is not None:
            child_overrides.update(self._overrides)
        child_overrides.update(overrides)

        retval = self.__class__(
            name=self.opt_name,
            model=self._model,
            algorithm_catalog=self._algorithm_catalog,
            generator=self._generator,
            overrides=child_overrides,
            **self._hyperparams
        )
        if self.queried:
            retval = retval.mark_queried()
        return retval



'''Base class for objects that perform varieties of cross validation.'''
# pylint: disable=too-many-arguments,duplicate-code
import abc
from typing import Any, Dict

from ...catalog.memory_catalog import MemoryCatalog
from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...executor.executor import Executor
from ...generator.bound_pipeline import BoundPipeline
from ...generator.designator import Designator
from ...instantiator.executable_pipeline import PipelineResult
from ...instantiator.instantiator_factory import InstantiatorFactory
from ...splitters.impl.splitter import SplitDataset

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class CrossValidatorError(Exception):
    '''Base error for cross-validators.'''


class CrossValidator(CatalogElementMixin, metaclass=abc.ABCMeta):
    '''Base class for objects that perform varieties of cross validation.'''

    _hyperparams: Dict[str, Any]

    def __init__(self, *unused_args, **default_hyperparams):
        super().__init__(unused_args, default_hyperparams)
        if not hasattr(self, '_hyperparams'):
            self._hyperparams = {}
        self._hyperparams.update(**default_hyperparams)

    def hyperparams(self, **overrides) -> Dict[str, Any]:
        '''Return cross-validator hyperparams, overridden with overrides'''
        retval = self._hyperparams.copy()
        retval.update(**overrides)
        return retval

    @abc.abstractmethod
    def validate_pipelines(self,
                           split_dataset: SplitDataset,
                           bound_pipelines: Dict[Designator, BoundPipeline],
                           instantiator: InstantiatorFactory,
                           executor: Executor,
                           **overrides
                           ) -> Dict[Designator, PipelineResult]:
        '''Do cross-validation by running pipelines on a split dataset.

        Returns the split dataset and an ExecutorResult for each pipeline.'''


class CrossValidatorCatalog(MemoryCatalog[CrossValidator], metaclass=abc.ABCMeta):
    '''Base class for cross-validator catalog'''



'''Tests for table_auto.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pandas as pd
import pytest

from ...wrangler.constants import PACKAGE_NAME

from .table import Table
from .table_auto import TableCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code


FAKE_TABLE_SOURCE = '''
import pandas as pd

from queue import Queue
from typing import Any, List, Optional
from ngautonml.tables.impl.table import TableFactory, Table
from ngautonml.tables.impl.table_catalog import TableCatalog
from ngautonml.tables.dataframe import DataFrameTable


class FakeType():
    def __init__(self, token: str = 'fake'):
        self._token = token


class FakeTable(Table):
    name = 'fake_table'
    tags = {
        'some_tag': ['some_value']
    }
    my_type = FakeType

    def to_csv(self, *args, **kwargs) -> str:
        return 'fake_table'

    def as_dict(self) -> dict:
        return {'fake_table': [1, 2, 3]}

    def to_list(self) -> list[int]:
        return [1, 2, 3]


def from_dataframe(table: DataFrameTable) -> FakeType:
    return table._as(pd.DataFrame).tolist()


def register(catalog: TableCatalog):
    catalog.register(FakeTable)
    TableFactory.register_constructor(FakeType, FakeTable)
    TableFactory.register_exporter(from_type=FakeType, to_type=list, exporter=FakeTable.to_list)
    TableFactory.register_exporter(from_type=pd.DataFrame, to_type=FakeType, exporter=from_dataframe)
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = (retval / f'.{PACKAGE_NAME}' / 'plugins'
                   / 'tables' / 'fake_table.py')
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_TABLE_SOURCE)
    return retval


def test_table_catalog_auto() -> None:
    dut = TableCatalogAuto()

    got = dut.lookup_by_name('dataframe')(
        value=pd.DataFrame(),
    )
    assert isinstance(got, Table)


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)

    dut = TableCatalogAuto()
    got = dut.lookup_by_name('fake_table')
    assert callable(got)
    # We need a FakeType object to create a FakeTable, so we're kind of stuck here.



'''Tests for time_since_send.py'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import time


from ..algorithms.impl.neighbor_metadata import NeighborMetadata
from ..algorithms.impl.toy_neighbor_state import ToyNeighborState
from ..config_components.distributed_config import DistributedConfig
from ..deciders.impl.decider import Decider
from ..neighbor_manager.node_id import NodeID

from .impl.decider_auto import DeciderCatalogAuto


def test_time_since_sent() -> None:
    '''Test the time_since_send decider with a geometric term.'''
    distributed = DistributedConfig(
        clause={
            'decider': {
                'time_since_sent': {
                    'time_seconds': 1.0,
                }
            },
        }
    )
    catalog = DeciderCatalogAuto()
    decider: Decider = catalog.lookup_by_name('time_since_sent')(
        config=distributed.decider.time_since_sent)  # type: ignore[attr-defined] # pylint: disable=no-member

    my_state = ToyNeighborState(center=0.0)
    my_meta = NeighborMetadata(my_state=my_state)

    neighbor_id = NodeID(-1)

    # We send the first time.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors={}) is True
    my_meta.sent_current_state()
    # We should not send immediately after sending...
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors={}) is False
    time.sleep(0.1)
    # ...or after a short interval.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors={}) is False
    time.sleep(1.0)
    # We should send after a full interval.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors={}) is True
    my_meta.sent_current_state()
    # See that the sequence repeats.
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors={}) is False
    time.sleep(0.1)
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors={}) is False
    time.sleep(1.0)
    assert decider.decide(my_meta=my_meta, neighbor_id=neighbor_id, neighbors={}) is True



"""The wrangler is the central control object for all of AutonML."""

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=too-many-instance-attributes,too-many-arguments

from typing import Dict, Optional

from ..aggregators.impl.aggregate_ranker import AggregateRanker
from ..generator.designator import Designator
from ..instantiator.executable_pipeline import (ExecutablePipeline,
                                                PipelineResult,
                                                PipelineResults)
from ..templates.impl.pipeline_template import PipelineTemplate

from .dataset import Dataset
from .logger import Level, Logger
from .wrangler_base import WranglerBase
from .wrangler_result import WranglerResult

log = Logger(__file__, level=Level.ERROR).logger()


class Wrangler(WranglerBase):
    '''The wrangler is the central control object for all of AutonML.'''

    def lookup_templates(self) -> Dict[str, PipelineTemplate]:
        '''Look up the templates that match the problem definition.'''
        task = self._pd.task
        data_type = "None"
        if task.data_type is not None:
            data_type = task.data_type.name
        task_type = "None"
        if task.task_type is not None:
            task_type = task.task_type.name

        return self._template_catalog.lookup_by_both(
            data_type=[data_type], task=[task_type])

    def train_all_results(self,
                          results: PipelineResults,
                          dataset: Dataset) -> PipelineResults:
        '''Instantiate and train all pipelines in results.

        Note that we do NOT calculate new predictions.
        '''
        retval = PipelineResults()
        pipelines: Dict[Designator, ExecutablePipeline] = {}

        for des, result in results.items():
            pipelines[des] = self._instantiator_factory.instantiate(
                kind=self._executor.kind,
                pipeline=result.bound_pipeline)

        self._executor.fit(
            dataset=dataset,
            pipelines=pipelines)

        for des, pipe in pipelines.items():
            retval[des] = PipelineResult(
                prediction=results[des].prediction,
                split_dataset=results[des].split_dataset,
                executable_pipeline=pipe)

        return retval

    def _predict_test_data(self,
                           executable_pipelines: Dict[Designator, ExecutablePipeline]
                           ) -> Optional[PipelineResults]:
        '''Run the executor again on test data to get test predictions,
            if test data is supplied in the problem definition.

        (Currently runs all pipelines, may eventually run x best ones)'''
        assert self._bound_pipelines is not None, (
            'BUG: wrangle_test() called with no bound pipelines.')

        test_dataset = self.load_test_dataset()

        if test_dataset is None:
            return None

        test_predictions = self._executor.predict(
            dataset=test_dataset,
            pipelines=executable_pipelines)
        return PipelineResults(test_predictions)

    def fit_predict_rank(self) -> WranglerResult:
        '''Do all the autoML things.'''
        train_dataset = self.load_train_dataset()

        self._bound_pipelines = self._build_pipelines()

        task = self._pd.task
        assert task.task_type is not None, (
            'BUG: missing task should have been caught in validation.')
        splitters = self._splitter_catalog.lookup_by_tag_and(**{
            'task': task.task_type.name,
            'data_type': task.data_type.name,
        })
        if not splitters:
            splitters = self._splitter_catalog.lookup_by_tag_and(**{
                'default': 'true'
            })
        assert len(splitters) == 1, f'BUG: More than one splitter returned for {task}: {splitters}'
        splitter = list(splitters.values())[0]

        assert train_dataset is not None
        split_data = splitter.split(
            dataset=train_dataset,
            **self._pd.cross_validation_config.splitter_hyperparams
        )
        log.info('Split dataset into %d folds.', len(split_data.folds))
        print(f'Split dataset into {len(split_data.folds)} folds.')

        cross_validator = self._validator_catalog.lookup_by_name('k_fold_cross_validator')
        cross_validator_results = PipelineResults(cross_validator.validate_pipelines(
            split_dataset=split_data,
            bound_pipelines=self._bound_pipelines,
            instantiator=self._instantiator_factory,
            executor=self._executor))

        metrics = self._metric_catalog.lookup_metrics(self._pd)
        print(f'Got {len(metrics)} metrics.')
        rankings = self._ranker.rank(
            results=cross_validator_results,
            metrics=metrics,
            ground_truth=split_data.ground_truth)
        log.info('Rankings: %s', [str(rank) for rank in rankings])

        methods = [self._aggregator_catalog.lookup_by_name(method)
                   for method in self._pd.aggregation.method]
        new_rankings = AggregateRanker(methods=methods,
                                       rankings=rankings,
                                       results=cross_validator_results)()
        rankings.update(new_rankings)
        print(f'Added {len(new_rankings)} aggregate rankings.')

        train_results = self.train_all_results(
            results=cross_validator_results,
            dataset=train_dataset)

        if self._all_executable_pipelines is None:
            self._all_executable_pipelines = train_results.executable_pipelines
            self._current_executable_pipelines = self._all_executable_pipelines

        if self._saver is not None:
            self.save(train_results)

        test_results = self._predict_test_data(train_results.executable_pipelines)

        return WranglerResult(
            split_dataset=split_data,
            train_results=train_results,
            test_results=test_results,
            rankings=rankings)



'''Base module for in memory catalogs.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from collections import defaultdict
from importlib import util
import os
from pathlib import Path
import sys
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, TypeVar, Union

from .catalog import (Catalog, CatalogError, CatalogLookupError,
                      CatalogNameError, CatalogValueError)
from ..wrangler.constants import CATALOG_IGNORE
from ..wrangler.logger import Level, Logger

log = Logger(__file__).logger()

T = TypeVar("T")


class CatalogNoRegister(CatalogError):
    '''A module lacks a register function.'''


class CatalogDuplicateError(CatalogError):
    '''We attempted to insert an element with a duplicate key.'''


class MemoryCatalog(Catalog[T]):
    '''In memory Generic Catalog class for managing swappable components.

    This version of Catalog allows registering an element and loading a directory.

    Allows search by name of the component, and a lookup
    by classification or property.
    '''
    _kwargs: Dict[str, Any]
    __cat_names: Dict[str, T]
    __tags: Dict[str, Dict[str, List[Tuple[str, T]]]]
    # "tag_name": {
    #     "tag_value"{
    #         [
    #             ("modelname1", Model1),
    #             ("modelname2", Model2)
    #         ]
    #     }
    # }
    __names_by_tag: Dict[str, Dict[str, List[str]]]
    # "tag_name": {
    #     "tag_value"{
    #         [
    #             "modelname1",
    #             "modelname2"
    #         ]
    #     }
    # }
    __tags_by_name: Dict[str, Dict[str, List[str]]]
    # "modelname1": {
    #     "tag_name"{
    #         [
    #             "tag_value1",
    #             "tag_value2"
    #         ]
    #     }
    # }

    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self.__cat_names = {}
        self.__tags = defaultdict(lambda: defaultdict(list))
        self.__names_by_tag = defaultdict(lambda: defaultdict(list))
        self.__tags_by_name = defaultdict(lambda: defaultdict(list))

    def register(self, obj: T, name: Optional[str] = None,
                 tags: Optional[Dict[str, Union[str, List[str]]]] = None) -> str:
        '''Register an object under the name.'''
        obj_name = name
        if obj_name is None:
            if hasattr(obj, 'name'):
                obj_name = getattr(obj, 'name')
            else:
                if isinstance(obj, type):
                    obj_signifier = obj.__name__
                else:
                    obj_signifier = obj.__class__.__name__

                raise CatalogNameError(
                    'Name must be defined for Catalog Object: '
                    f'{obj_signifier}')
        obj_name = obj_name.lower()

        if obj_name in self.__cat_names:
            raise CatalogDuplicateError(obj_name)

        self.__cat_names[obj_name] = obj

        if tags is None:
            tags = getattr(obj, 'tags', None)

        if tags is not None:
            for ttype, tvals in tags.items():
                if isinstance(tvals, str):
                    tvals = [tvals]
                for tval in tvals:
                    if not isinstance(tval, str):
                        raise CatalogValueError(f'{ttype}:{tval} is a {type(tval)}, not string')
                    self.__tags[ttype.lower()][tval.lower()].append((obj_name, obj))
                    self.__names_by_tag[ttype.lower()][tval.lower()].append(obj_name)
                    self.__tags_by_name[obj_name][ttype.lower()].append(tval.lower())
        return obj_name

    @property
    def tagtypes(self) -> Set[str]:
        '''Retrieve list of all tag types present in catalog'''
        return set(self.__tags.keys())

    def all_objects(self) -> Iterable[T]:
        '''Retrieve all registered objects.'''
        return self.__cat_names.values()

    def tagvals(self, tagtype: str) -> Set[str]:
        '''Retrieve all tag values present for a given tag type'''
        if tagtype in self.__tags.keys():
            return set(self.__tags[tagtype].keys())
        return set()

    def lookup_by_name(self, name: str) -> T:
        '''Find an object registered under name.'''
        obj_name = name.lower()
        if obj_name in self.__cat_names:
            return self.__cat_names[obj_name]

        raise CatalogLookupError(name)

    def lookup_by_tag_and(self, **tags: Union[str, Iterable[str]]) -> Dict[str, T]:
        '''Find objects with a specified tag for all specified tag types.

        \\*\\*tags looks like:
            tag_type = tag_value
            OR
            tag_type = (tag_value1, tag_value2)

        Needs to match at least one specified tag value for all specified tag types.
        '''
        result: Dict[str, T] = self.__cat_names.copy()
        for tag_type, tag_values in tags.items():
            if isinstance(tag_values, str):
                tag_values = (tag_values,)

            result = {
                name: obj
                for name, obj in result.items()
                if name in self._union_tag_search(tag_type, tag_values)
            }

        return result

    def _union_tag_search(self,
                          tag_type: str,
                          tag_values: Iterable[str]) -> List[str]:
        '''Return all the names in the catalog that match at least 1 provided tag value'''
        found_names: Set[str] = set()
        tag_type = tag_type.lower()
        for tag_value in tag_values:
            tag_value = tag_value.lower()
            found_names.update(set(self.__names_by_tag[tag_type][tag_value]))
        return list(found_names)

    def load(self, module_directory: Path):
        '''Load all the catalog modules from a directory hierarchy.

        Normal usage is to set module_directory to Path(__file__).parents[1].

        Each module in the hierarchy needs a "register(catalog, \\*arg, \\*\\*kwargs)"
        function that will register all the catalog objects in the module.
        '''
        module_directory = module_directory.resolve()
        root = python_root(str(module_directory))
        filepaths = []

        # Build up the list of files to examine.
        for rootdir, dirs, filenames in os.walk(module_directory, topdown=True):
            if CATALOG_IGNORE in dirs:
                log.log(Level.VERBOSE, 'dropping %s/%s', rootdir, CATALOG_IGNORE)
                dirs[:] = [d for d in dirs if d != CATALOG_IGNORE]
            for filename in filenames:
                if self._skip(rootdir, filename):
                    continue
                filepaths.append(os.path.abspath(os.path.join(rootdir, filename)))
        # Load all the modules calling register (if present).
        for filepath in filepaths:
            modname = module_name_from_filepath(filepath, root)

            module = self._load_module(filepath, modname)

            # At this point we think we have a valid module.
            if hasattr(module, 'register'):
                module.register(self, **self._kwargs)
            else:
                raise CatalogNoRegister(
                    f'autoloader: module {module.__name__} '
                    f'has no register, dir(module): {dir(module)}')

    def _skip(self, rootdir, filename) -> bool:
        '''Returns True if the autoloader should skip this file, and logs it.'''
        if not filename.endswith('.py'):
            log.log(Level.VERBOSE, 'autoloader dropping non-python file %s/%s', rootdir, filename)
            return True
        if filename.startswith('__'):
            log.log(Level.VERBOSE, 'autoloader dropping special file %s/%s', rootdir, filename)
            return True
        if filename.endswith('_test.py'):
            log.log(Level.VERBOSE, 'autoloader dropping test path %s/%s', rootdir, filename)
            return True
        return False

    def _load_module(self, filepath, modname):
        '''Load the specified file as a module.'''
        if modname in sys.modules:
            # If module is already loaded: don't reload it.
            log.log(Level.VERBOSE, '%s is already loaded', modname)
            module = sys.modules[modname]

        else:
            # Load the file as a module.
            spec = util.spec_from_file_location(modname, filepath)
            if spec is None:
                raise ImportError(f"can't make spec for module {modname} at {filepath}")
            module = util.module_from_spec(spec)
            if spec.loader is None:
                raise ImportError(f'loader for module {modname} is None')
            spec.loader.exec_module(module)
            sys.modules[modname] = module
        return module

    def items(self) -> Iterable[Tuple[str, T]]:
        return self.__cat_names.items()


def module_name_from_filepath(filepath: str, root_path: str) -> str:
    """Converts a path to a .py file into a Python module name.

    E.g. for a root_path of "/home/piggy/src/autonml" and a filepath of
    "/home/piggy/src/autonml/ngautonml/algorithms/connect.py", we get
    a module name of "ngautonml.algorithms.connect".

    Args:
        filepath: The path to the .py file.
        root_path: The path to the root of the Python tree that the module lives in.

    Returns:
        The full name of the module pointed at by filepath.
    """
    filepath = os.path.abspath(filepath)
    root_path = os.path.abspath(root_path)

    # Get the relative path to the .py file from the root of the Python tree.
    relative_path = os.path.relpath(path=filepath, start=root_path)

    # Split the relative path on the last "/".
    path_components = relative_path.split(os.sep)

    # Strip off the .py extension from the last component of the path, if it exists.
    if path_components[-1].endswith(".py"):
        path_components[-1] = path_components[-1][:-3]

    # Join the path components with dots.
    module_name = ".".join(path_components)

    return module_name


def python_root(filepath: str) -> str:
    """Find the top directory with an __init__.py in it.

    Args:
        filepath: The path to a Python file.

    Returns:
        The prefix that points to the highest directory with an __init__.py.
    """

    filepath = os.path.abspath(filepath)
    path_components = filepath.split(os.sep)

    # Find the first directory with an __init__.py in it.
    for i in range(len(path_components)):
        if os.path.exists(os.path.join(*path_components[:i], "__init__.py")):
            break

    # Remove all directories before the directory with an __init__.py from the path.
    # Please don't put __init__.py in /.
    path_components = path_components[:i - 1]

    # Join the path components with os.path.join.
    prefix = os.path.sep + os.path.join(*path_components)

    return prefix



'''Specified dataset handling'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring

from typing import Any, Dict, List, Optional

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..problem_def.problem_def_interface import ProblemDefInterface
from ..wrangler.constants import ProblemDefKeys, ProblemDefKeySet, Defaults
from ..wrangler.dataset import Column, Metadata, RoleName
from .impl.config_component import (
    ConfigComponent, ConfigError, InvalidKeyError, InvalidValueError,
    MissingKeyError, ValidationErrors)
from .impl.config_component_catalog import ConfigComponentCatalog


class Error(ConfigError):
    '''Base class for errors in this file.'''


class UnknownConfigClassError(Error):
    '''Problem definition specified a dataset config that is not recognized.'''


class DatasetFileError(Error):
    '''The dataset file does not match information given in the problem definition.'''


class DatasetConfigTypeError(Error, TypeError):
    '''An object in the dataset config was not the right type.'''


DEFAULT_LOADED_FORMATS = {
    'tabular_file': 'pandas_dataframe',
    'pandas_dataframe': 'pandas_dataframe',
    'image_dir': 'tensorflow'
}


class DatasetConfig(ConfigComponent):
    '''Specifies dataset handling.'''
    name = 'dataset'
    tags: Dict[str, Any] = {}

    class Keys(AEnum):
        '''Standard strings used for keys of this clause'''
        CONFIG = 'config'
        COL_ROLES = 'column_roles'
        TRAIN_PATH = 'train_path'
        TEST_PATH = 'test_path'
        TRAIN_DIR = 'train_dir'  # used for ImageDirDataLoader
        TEST_DIR = 'test_dir'  # used for ImageDirDataLoader
        TRAIN_DATA = 'train_data'  # Will eventually be all-purpose (obj, csv, dir, url)
        TEST_DATA = 'test_data'  # Will eventually be all-purpose (obj, csv, dir, url)
        STATIC_EXOGENOUS_PATH = 'static_exogenous_path'  # For forecasting problems only
        INPUT_FORMAT = 'input_format'
        LOADED_FORMAT = 'loaded_format'
        PARAMS = 'params'

    def __init__(self,
                 clause: Dict[str, Any],
                 problem_def: Optional[ProblemDefInterface] = None,
                 **unused_kwargs):
        parents = self._add_parent([], ProblemDefKeys.DATASET.value)
        super().__init__(clause=clause, parents=parents)
        self._roles = self._build_roles()
        self._pos_labels = self._build_pos_labels()
        self._problem_def = problem_def

    def _build_roles(self) -> Dict[RoleName, List[Column]]:
        '''Build dict of column roles'''
        retval: Dict[RoleName, List[Column]] = {}
        for role in self._get_with_default(self.Keys.COL_ROLES, dflt={}):
            # this will throw a parsing error if the key does not match any known role name
            rolename = RoleName[role.upper()]

            col_names = self._get_with_default(self.Keys.COL_ROLES,
                                               role,
                                               ProblemDefKeys.COL_NAME,
                                               dflt=None)
            # TODO(Merritt): don't use ProblemDefKeys here
            if col_names is not None:
                retval[rolename] = [Column(name=col_names)]
            # TODO(Merritt): this assumes one col per role, modify it

        return retval

    def _build_pos_labels(self) -> Dict[RoleName, Any]:
        '''Build dict of pos labels for binary classification.

        In the vast majority of cases, the only key will be RoleName.TARGET'''
        retval: Dict[RoleName, Any] = {}
        for role in self._get_with_default(self.Keys.COL_ROLES, dflt={}):
            rolename = RoleName[role.upper()]
            retval[rolename] = self._get_with_default(
                self.Keys.COL_ROLES, role, ProblemDefKeys.POS_LABEL, dflt=None)
            # TODO(Merritt): don't use ProblemDefKeys here
        return retval

    def validate(self, **kwargs) -> None:
        '''Check the problem definition dataset clause for errors.'''

        # fatal errors

        dataset_keys = set(self._clause.keys())
        if not self.required_keys().issubset(dataset_keys):
            raise ValidationErrors(errors=[MissingKeyError(
                'Required keys missing in dataset clause: '
                f'{self.required_keys().difference(dataset_keys)}')])

        # non-fatal errors

        errors: List[ConfigError] = []

        if not dataset_keys.issubset(self.allowed_keys()):
            errors.append(InvalidKeyError(
                'Invalid key(s) in new dataset clause: '
                f'{dataset_keys.difference(self.allowed_keys())} '
                f'Valid keys are {self.allowed_keys()}'))

        # TODO(piggy/Merritt) Move this to the data loader(s) that require
        # column roles.
        # If no col_roles clause exists, skip this validation.
        for role_name in self._get_with_default(self.Keys.COL_ROLES, dflt=[]):
            role_keys = set(self._get(self.Keys.COL_ROLES, role_name).keys())
            if not ProblemDefKeySet.DATASET.ROLES.REQUIRED.issubset(role_keys):
                errors.append(MissingKeyError(
                    f'Required key(s) missing in role clause for role {role_name}: '
                    f'{ProblemDefKeySet.DATASET.ROLES.REQUIRED.difference(role_keys)}'))

            if not role_keys.issubset(ProblemDefKeySet.DATASET.ROLES.ALLOWED):
                errors.append(InvalidKeyError(
                    f'Invalid key(s) in role clause for role {role_name}: '
                    f'{role_keys.difference(ProblemDefKeySet.DATASET.ROLES.ALLOWED)}'
                ))

        # 'params' subclause must be a dict
        if self._exists(self.Keys.PARAMS):
            params_clause = self._get(self.Keys.PARAMS)
            if not isinstance(params_clause, Dict):
                errors.append(InvalidValueError(
                    f'Expected dict at {ProblemDefKeys.DATASET.value}'  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
                    f'.{self.Keys.PARAMS.value}, instead found '
                    f'{type(params_clause)}.'
                ))

        if len(errors) > 0:
            raise ValidationErrors(errors=errors)

    @property
    def dataloader_tags(self) -> Dict[str, str]:
        '''These tags select the data loader.

        input_format is the format being loaded:
            tabular_file, pandas_dataframe, image directory.

        loaded_format is the format passed to subsequent stages:
            pandas_dataframe or tensorflow tensor
        '''
        input_format = self._get(self.Keys.INPUT_FORMAT)
        if self._exists(self.Keys.LOADED_FORMAT):
            loaded_format = self._get(self.Keys.LOADED_FORMAT)
        else:
            # We deliberately avoid _get_with_default so that we
            # only do this lookup if loaded_format is not specified.
            loaded_format = DEFAULT_LOADED_FORMATS.get(
                input_format, Defaults.LOADED_FORMAT)

        return {
            self.Keys.INPUT_FORMAT.value: input_format,  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            self.Keys.LOADED_FORMAT.value: loaded_format,  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        }

    @property
    def config(self) -> Optional[str]:
        '''The name of the data loader for this dataset.'''
        return self._get_with_default(self.Keys.CONFIG, dflt=None)

    @property
    def params(self) -> Dict[str, Any]:
        default_params = {}
        # Implement backwards compatability for sytax with params clause.
        if self._exists(self.Keys.TRAIN_PATH):
            default_params[self.Keys.TRAIN_PATH.value] = self._get(self.Keys.TRAIN_PATH)  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        if self._exists(self.Keys.TEST_PATH):
            default_params[self.Keys.TEST_PATH.value] = self._get(self.Keys.TEST_PATH)  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        if self._exists(self.Keys.STATIC_EXOGENOUS_PATH):
            default_params[self.Keys.STATIC_EXOGENOUS_PATH.value] = (  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
                self._get(self.Keys.STATIC_EXOGENOUS_PATH))
        if self._exists(self.Keys.TRAIN_DIR):
            default_params[self.Keys.TRAIN_DIR.value] = self._get(self.Keys.TRAIN_DIR)  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        if self._exists(self.Keys.TEST_DIR):
            default_params[self.Keys.TEST_DIR.value] = self._get(self.Keys.TEST_DIR)  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long

        retval = self._get_with_default(
            self.Keys.PARAMS, dflt=default_params)

        assert isinstance(retval, Dict)
        return retval

    @property
    def metadata(self) -> Metadata:
        '''The metadata needed by models.'''

        return Metadata(problem_def=self._problem_def,
                        roles=self._roles,
                        pos_labels=self._pos_labels)


def register(catalog: ConfigComponentCatalog) -> None:
    '''Register all the objects in this file.'''
    catalog.register(DatasetConfig)



'''A model that transforms time series forecasting data from wide format to long format.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional
import pandas as pd

from ..algorithms.impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..algorithms.impl.algorithm_instance import AlgorithmInstance
from ..catalog.catalog import upcast
from ..wrangler.dataset import Column, Dataset, RoleName


class WideToLongModelInstance(AlgorithmInstance):
    '''A model that transforms time series forecasting data from wide format to long format.'''

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        '''Transform time series dataset from wide to long format.'''
        if dataset is None:
            return None

        input_roles = dataset.metadata.roles

        # Ensure the dataset has exactly 1 time column and at least 1 target column.
        assert RoleName.TIME in input_roles, (
            'BUG: input to WideToLongModel has no time column. \n'
            f'Roles: {input_roles}'
        )
        assert 1 == len(input_roles[RoleName.TIME]), (
            'BUG: input to WideToLongModel must have exactly 1 time column. '
            f'Roles: {input_roles}'
        )
        assert RoleName.TARGET in input_roles, (
            'BUG: input to WideToLongModel has no target column. \n'
            f'Roles: {input_roles}'
        )

        # If dataset is already in long format, return a copy of it.
        if RoleName.TIMESERIES_ID in input_roles:

            assert len(input_roles[RoleName.TIMESERIES_ID]) == 1, (
                'BUG: dataset has more than 1 grouping key. \n'
                f'Roles: {input_roles}'
            )
            assert len(input_roles[RoleName.TARGET]) == 1, (
                'BUG: dataset has a grouping key but multiple targets. \n'
                f'Roles: {input_roles}'
            )
            copy_dataset = dataset.output()
            copy_dataset.dataframe = dataset.dataframe
            return copy_dataset

        if RoleName.PAST_EXOGENOUS in input_roles or RoleName.FUTURE_EXOGENOUS in input_roles:
            raise NotImplementedError(
                'Cannot transform dataset from wide to long format with exogenous variables. \n'
                f'Roles: {input_roles}')

        olddf = dataset.get_dataframe()
        newdf = pd.melt(
            olddf,
            id_vars=[str(col.name) for col in dataset.roles[RoleName.TIME]],
            var_name='unique_id',
            value_name='y'
        )

        roles = dataset.metadata.roles.copy()
        roles.update({
            RoleName.TIMESERIES_ID: [Column('unique_id')],
            RoleName.TARGET: [Column('y')]
        })

        new_dataset = dataset.output(
            override_metadata=dataset.metadata.override_roles(roles=roles)
        )
        new_dataset.dataframe = newdf
        return new_dataset


class WideToLongModel(Algorithm):
    '''A model that transforms time series forecasting data from wide format to long format.

    If input data is already long, it will return output as input.

    Wide format means multiple time series are in different columns, such as:
    time   ts1     ts2
    1      1.1     2.1
    2      1.2     2.2
    3      1.3     2.3

    Long format means multiple time series are in the same column,
    distinguished by a grouping key.  An example:
    time   key     value
    1      ts1     1.1
    2      ts1     1.2
    3      ts1     1.3
    1      ts2     2.1
    2      ts2     2.2
    3      ts3     2.3
    '''
    _name = 'wide_to_long'
    _instance_constructor = WideToLongModelInstance
    _tags = {
        'source': ['pandas'],
        'preprocessor': ['true'],
    }


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = WideToLongModel(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Metric configuration'''
from enum import Enum
from typing import Any, Dict, List, Optional, Set

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..config_components.impl.config_component import ConfigComponent, ConfigError


class MetricConfigError(ConfigError):
    '''Base error for MetricConfig.'''


class MetricConfig(ConfigComponent):
    '''Configured metrics'''
    _catalog_name: Optional[str] = None

    class Constants(Enum):
        '''Constants below and above top level.'''
        CATALOG_NAME = 'catalog_name'
        METRICS = 'metrics'

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        parents = self._add_parent(parents, self.Constants.METRICS.value)
        super().__init__(clause=clause, parents=parents)

    def validate(self, **kwargs) -> None:
        '''Throw a MetricConfigError if the metric clause is invalid'''
        errors: List[MetricConfigError] = []

        if len(self._clause.keys()) != 1:
            errors.append(MetricConfigError(
                'Metric clause must have exactly 1 key;'
                f'found {len(self._clause.keys())} ({self._clause.keys()})'
            ))

        try:
            # These are all run to make sure they generate no errors.
            self.name  # pylint: disable=pointless-statement
            self.hyperparams  # pylint: disable=pointless-statement
            self.catalog_name  # pylint: disable=pointless-statement
        except AssertionError as err:
            errors.append(MetricConfigError(str(err)))

        if len(errors) > 0:
            raise MetricConfigError(errors)

    def allowed_keys(self) -> Set[str]:
        return set(self._clause.keys())

    @property
    def name(self) -> str:
        '''The instance name for this metric.'''
        return list(self._clause.keys())[0]

    @property
    def hyperparams(self) -> Dict[str, Any]:
        '''The hyperparams for this metric.

        If you want more than one instance of a single metric
        with different hyperparams, define each with its own
        name and use the 'catalog_name' to specify the metric
        in the catalog to use.
        '''
        retval = dict(list(self._clause.values())[0])
        if self.Constants.CATALOG_NAME.value in retval:
            del retval[self.Constants.CATALOG_NAME.value]
        return retval

    @property
    def catalog_name(self) -> str:
        '''The metric catalog name for this metric.

        We can have different metric entries with the same catalog_name,
        but different hyperparams.

        Defaults to self.name.
        '''
        metric_info: Dict[str, Any] = list(self._clause.values())[0]
        retval = metric_info.get(self.Constants.CATALOG_NAME.value, self.name)
        assert isinstance(retval, str), f'catalog_name must be a str, not a {type(retval)}'
        return retval



'''The simplest abstractor algorithm that just shares data.'''

from typing import Dict, Iterable, List, Optional

import pickle

import pandas as pd

from ...catalog.catalog import upcast
from ...wrangler.dataset import Dataset
from ...wrangler.logger import Logger, Level

from ..impl.algorithm import MemoryAlgorithmCatalog

from .impl.abstractor import Abstractor
from .impl.abstractor_instance import (
    AbstractorInstance, AbstractorNeighborState)

log = Logger(__file__, to_stdout=False, level=Level.DEBUG).logger()

# pylint: disable=protected-access


class ShareAbstractor(Abstractor):
    '''An abstractor that just copies data.'''
    _name = 'share_abstractor'
    _default_hyperparams = {  # pylint: disable=duplicate-code
        'Lambda': 1.0,
        'omega': 0.667,
        'tol': 0.000000001,
        'maxiter': None,
    }
    _tags: Dict[str, List[str]] = {
        'source': ['autonlab'],
        'distributed': ['true'],
        'abstractor': ['true'],
    }

    def instantiate(self, **hyperparams) -> 'ShareAbstractorInstance':
        return ShareAbstractorInstance(parent=self, **hyperparams)


class ShareAbstractorNeighborState(AbstractorNeighborState):
    '''A abstractor neighbor state that just holds data.'''
    _data: Dataset

    def __init__(self, columns: list, data: Dataset):
        # Remove duplicate rows to avoid exponential growth.
        unique_data = data.dataframe.drop_duplicates()
        self._data = data.output()
        self._data.dataframe = unique_data
        super().__init__(columns=columns)
        # TODO(piggy): A much more network-efficient way to do this would be to
        # keep track of our net dataset separately, and only communicate the
        # diff between our dataset and the union of all our neighbors' datasets.
        # We are converged when the diff is empty.

    def encode(self) -> bytes:
        '''Encode a message for distributed neighbors. '''
        return pickle.dumps((self._columns, self._data))

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'ShareAbstractorNeighborState':
        '''Decode a message from distributed neighbors.'''
        columns, data = pickle.loads(serialized_model)
        return cls(columns=columns, data=data)

    def distance(self, other: Optional['ShareAbstractorNeighborState']) -> float:
        '''Return the distance to another state.'''
        if other is None or self._data is None or other._data is None:
            return 1.0
        if set(self._data.dataframe.columns) != set(other._data.dataframe.columns):
            return 1.0
        if len(self._data.dataframe) != len(other._data.dataframe):
            return 1.0
        comparison = self._data.dataframe.compare(other._data.dataframe)
        return 1.0 * comparison.size

    def synthesize(self) -> Dataset:
        '''Return the data we have.'''
        return self._data


class ShareAbstractorInstance(AbstractorInstance):
    '''An instance of AbstractorInstance that just shares data.'''
    _neighbor_constructor = ShareAbstractorNeighborState

    @property
    def _neighbor_models_iter(self) -> Iterable[ShareAbstractorNeighborState]:
        for v in super()._neighbor_models_iter:
            assert isinstance(v, ShareAbstractorNeighborState), (
                'BUG: expected neighbor_models to contain ShareAbstractorNeighborState, '
                f'instead found {v} of type {type(v)}'
            )
            yield v

    def __init__(self, parent, distributed, **kwargs):
        self.hyperparams = parent.hyperparams(**kwargs)
        super().__init__(parent=parent, distributed=distributed, **parent.hyperparams(**kwargs))

    def _decode(self, serialized_model: bytes) -> ShareAbstractorNeighborState:
        '''Decode a message from distributed neighbors. '''
        return ShareAbstractorNeighborState.decode(serialized_model)

    def _fit(self, dataset: Optional[Dataset], **kwargs) -> None:
        '''Concatenate, sort, and uniquify.
        '''
        if dataset is None and not self._neighbor_metadata:
            return
        new_data: Dataset
        new_df = pd.DataFrame()
        if self._neighbor_metadata:
            for neighbor in self._neighbor_models_iter:
                new_df = pd.concat([new_df, neighbor._data.dataframe])
            new_data = next(self._neighbor_models_iter)._data.output()  # type: ignore
        if dataset is not None:
            new_df = pd.concat([new_df, dataset.dataframe])
            new_data = dataset.output()

        new_df = new_df.sort_values(by=list(new_df.columns))
        new_df = new_df.drop_duplicates()

        new_df.reset_index(drop=True, inplace=True)

        new_data.dataframe = new_df

        self._my_state = ShareAbstractorNeighborState(columns=list(new_df.columns), data=new_data)

    @property
    def my_state(self) -> Optional[ShareAbstractorNeighborState]:
        '''Accessor so that tests can see the state'''
        retval = self._my_state
        if retval is not None:
            assert isinstance(retval, ShareAbstractorNeighborState)
        return retval


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = ShareAbstractor(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Identifies nodes in a distributed AI setting'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


class NodeID(int):
    '''Each neighbor is uniquely identified by an integer.'''



'''Tests for linear_distributed_algorithm_instance.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code

import numpy as np

from ...config_components.distributed_config import DistributedConfig

from .linear_distributed_algorithm_instance import (
    LinearNeighborState)


class FakeLinearNeighborState(LinearNeighborState):
    '''Linear neighbor state that does not care about encode() and decode()'''

    def encode(self) -> bytes:
        '''Encode message for distributed neighbors.'''
        return b'fake message'

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'FakeLinearNeighborState':
        '''Decode a message from a neighbor.'''
        return FakeLinearNeighborState(np.array([1, 2, 3]))


def test_should_send() -> None:
    '''_should_send returns True iff our model changed.'''

    config = DistributedConfig({})
    dut = FakeLinearNeighborState(v=np.asarray([1.0, 2.0, 3.0]))
    same_as_dut = FakeLinearNeighborState(v=np.asarray([1.0, 2.0, 3.0]))
    different_from_dut = FakeLinearNeighborState(v=np.asarray([0.1, 0.2, 0.3]))

    assert dut.state_differs(
        distributed=config,
        other=different_from_dut
    ) is True

    assert dut.state_differs(
        distributed=config,
        other=same_as_dut
    ) is False



'''tests for ssnm.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from ..config_components.distributed_config import DistributedConfig
from ..communicators.stub_communicator import CommunicatorStub
from ..discoverers.static_discoverer import StaticDiscoverer

from .node_id import NodeID
from .neighbor_manager import NewNeighbor
from .neighbor_manager_impl import NeighborManagerImpl

# pylint: disable=missing-function-docstring, duplicate-code


def test_neighbor_startup():
    distributed_config = DistributedConfig(clause={
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2, 4, 5],
                    '2': [1, 3, 5],
                    '3': [5, 2],
                    '4': [1],
                    '5': [1, 2, 3]
                }
            }
        }
    })
    my_id = NodeID(2)
    discoverer = StaticDiscoverer(
        config=distributed_config,
        communicator=CommunicatorStub(my_id=my_id))
    dut = NeighborManagerImpl(discoverer=discoverer)

    dut.start()

    # WARNING: By the rules for queue.Queue this test
    # could be flakey. We're not guaranteed to be able
    # to read all the queued NewNeighbor events.
    events = dut.poll_for_events()

    assert all(isinstance(e, NewNeighbor) for e in events)

    neighbors = set(e.neighbor for e in events)

    assert neighbors == {NodeID(1), NodeID(3), NodeID(5)}

    dut.stop()



'''Tests for pathed_catalog.py.'''

from pathlib import Path

import pytest

from .catalog import Catalog
from .pathed_catalog import PathedCatalog
from .widgets.impl.widget_catalog import NamedWidget

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code


EXTRA_WIDGET_SOURCE = '''
from typing import Any, Optional
from ngautonml.catalog.widgets.impl.widget_catalog import NamedWidget, WidgetCatalog

class ExtraWidget(NamedWidget):
    _name = 'extra_widget'
    _tags = {
        'some_tag': ['extra_dir']
    }


def register(catalog: WidgetCatalog):
    catalog.register(ExtraWidget())
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = retval / 'extra_widget.py'
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(EXTRA_WIDGET_SOURCE)
    return retval


def test_home_directory_auto(tmp_path) -> None:
    dut: Catalog = PathedCatalog(paths=[tmp_path, Path(__file__).parent / 'widgets'])
    got1 = dut.lookup_by_name('extra_widget')
    assert isinstance(got1, NamedWidget)
    assert got1.tags['some_tag'] == ['extra_dir']
    got2 = dut.lookup_by_name('Gizmo')
    assert isinstance(got2, NamedWidget)
    assert got2.tags['some_tag'] == ['test_dir']
    assert hasattr(got2, 'quux')



'''Holds configuration information for cross-validation.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
from typing import Dict, Any, Optional, List

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..config_components.impl.config_component import ConfigComponent, InvalidValueError
from ..wrangler.constants import ProblemDefKeys, Defaults


class CrossValidationConfig(ConfigComponent):
    '''
    Holds configuration information for cross-validation.
    '''

    def __init__(self, clause: Dict[str, Any], parents: Optional[List[str]] = None):
        parents = self._add_parent(parents, ProblemDefKeys.CROSS_VALIDATION.value)
        super().__init__(clause=clause, parents=parents)

    class Constants(Enum):
        '''Lower-level clauses in cross_validation component'''
        N_SPLITS = 'n_splits'

    class Keys(AEnum):
        '''Top-level clauses in cross_validation component'''
        K = 'k'
        SEED = 'seed'

    @property
    def k(self) -> Optional[int]:
        '''get k fold'''
        field = self._get_with_default(self.Keys.K, dflt=None)

        return field

    @property
    def seed(self) -> int:
        '''Get default random seed or one set by'''
        field = self._get_with_default(
            self.Keys.SEED,
            dflt=Defaults.SEED
        )
        return field

    @property
    def splitter_hyperparams(self) -> Dict[str, Any]:
        '''get splitter hyperparams'''
        retval = {}

        if self.k:
            retval[self.Constants.N_SPLITS.value] = self.k

        return retval

    def validate(self, **kwargs) -> None:
        if self.k is None:
            return

        # single fold cross-validation does not make sense
        #   because we need data outside the fold to train on
        if self.k < 2:
            raise InvalidValueError(
                f'k must be an integer greater than 1, but we got {self.k} instead.')

        df_row_count = 'df_row_count'

        if df_row_count not in kwargs:
            return

        count = kwargs[df_row_count]

        if self.k > count:
            raise InvalidValueError(
                f'k must not be greater than the total number of rows in the dataframe.'
                f'Instead we got: k = {self.k}, row count = {count}'
            )



'''Wraps test aggregators with TSADAggregator and registers them.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.


from typing import Any, Dict, List
from .impl.aggregator_catalog import AggregatorCatalog
from .tsad_aggregator import TSADAggregator


AGGREGATORS: List[Dict[str, Any]] = [
    {
        'name': 'test_aggregators.identity',
        'tags': {'for_tests': ['true']},
    },
]


def register(catalog: AggregatorCatalog, *unused_args, **unused_kwargs) -> None:
    '''Register all the objects in this file.'''
    for agg in AGGREGATORS:
        catalog.register(TSADAggregator(**agg))



'''Test the catalog module.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path

import pytest
from .memory_catalog import CatalogDuplicateError
from .catalog_element_mixin import CatalogElementMixin
from .widgets.impl.widget_catalog import NamedWidget, WidgetCatalog, Widget
# pylint: disable=missing-function-docstring,missing-class-docstring,duplicate-code


def test_register() -> None:
    '''Test register with an implicit name.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    assert dut.register(test_widget) == 'testwidget'


def test_register_with_name() -> None:
    '''Test register, specifying a name.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    assert dut.register(test_widget, 'TestName') == 'testname'


def test_register_duplicate() -> None:
    '''Test duplicate registration.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    dut.register(test_widget, 'TestName')
    with pytest.raises(CatalogDuplicateError):
        dut.register(test_widget, 'TestName')


def test_lookup_by_name() -> None:
    '''Sunny day test of lookup_by_name().'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    dut.register(test_widget, 'TestName')
    assert dut.lookup_by_name('TestName') == test_widget


def test_lookup_one_by_tag() -> None:
    '''Find a list of one widget by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    dut.register(test_widget, 'TestWidget', tags={'userdef': ['TestTag']})
    assert len(dut.lookup_by_tag_and(userdef='TestTag')) == 1


def test_lookup_one_by_tag_case_difference() -> None:
    '''Find a list of one widget by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    dut.register(test_widget, 'TestWidget', tags={'userdef': ['TestTag']})
    assert len(dut.lookup_by_tag_and(userdef='testtag')) == 1


def test_lookup_one_by_tag_case_difference_type() -> None:
    '''Find a list of one widget by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    dut.register(test_widget, 'TestWidget', tags={'UserDef': ['TestTag']})
    assert len(dut.lookup_by_tag_and(userdef='TestTag')) == 1


def test_lookup_multi_by_tag() -> None:
    '''Find a list of several widgets by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widgetone = NamedWidget('TestWidget')
    dut.register(test_widgetone, 'TestWidget', tags={'userdef': ['TestTag']})
    test_widgettwo = NamedWidget('TestWidgetTwo')
    dut.register(test_widgettwo, 'TestWidgetTwo', tags={'userdef': ['TestTag']})
    test_widgetthree = NamedWidget('TestWidgetThree')
    dut.register(test_widgetthree, 'TestWidgetThree', tags={'userdef': ['TestTag']})
    test_widgetfour = NamedWidget('TestWidgetFour')
    dut.register(test_widgetfour, 'TestWidgetFour', tags={'userdef': ['AnotherValue']})

    assert len(dut.lookup_by_tag_and(userdef='TestTag')) == 3


def test_lookup_multi_by_tag_type_diff() -> None:
    '''Find a list of several widgets by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widgetone = NamedWidget('TestWidget')
    dut.register(test_widgetone, 'TestWidget', tags={'userdef': ['TestTag']})
    test_widgettwo = NamedWidget('TestWidgetTwo')
    dut.register(test_widgettwo, 'TestWidgetTwo', tags={'userdef': ['TestTag']})
    test_widgetthree = NamedWidget('TestWidgetThree')
    dut.register(test_widgetthree, 'TestWidgetThree', tags={'userdef': ['TestTag']})
    test_widgetfour = NamedWidget('TestWidgetFour')
    dut.register(test_widgetfour, 'TestWidgetFour', tags={'anothertype': ['TestTag']})

    assert len(dut.lookup_by_tag_and(userdef='TestTag')) == 3


def test_lookup_by_tag_not_found() -> None:
    '''Fail to find by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    dut.register(test_widget, 'TestWidget')
    assert len(dut.lookup_by_tag_and(userdef='TestTag')) == 0


def test_lookup_by_tagtype_not_found() -> None:
    '''Fail to find by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    dut.register(test_widget, 'TestWidget', tags={'userdef': ['TestTag']})
    assert len(dut.lookup_by_tag_and(anothertype='TestTag')) == 0


def test_get_tagtypes() -> None:
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    test_widget2 = NamedWidget('TestWidget2')
    assert not dut.tagtypes
    dut.register(test_widget, tags={'k1': ['a']})
    dut.register(test_widget2, tags={'k1': ['b'], 'k2': ['c']})
    assert dut.tagtypes == {'k1', 'k2'}


def test_get_tagvals() -> None:
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    test_widget2 = NamedWidget('TestWidget2')
    dut.register(test_widget, tags={'k1': ['a']})
    dut.register(test_widget2, tags={'k1': ['b'], 'k2': ['c']})
    assert dut.tagvals('k1') == {'a', 'b'}


def test_lookup_multiple_by_mutiple_tag() -> None:
    '''Find a list of multiple widgets by matching multiple tags'''
    dut = WidgetCatalog()
    test_widget1 = NamedWidget('TestWidget1')
    test_widget2 = NamedWidget('TestWidget2')
    test_widget3 = NamedWidget('TestWidget3')
    dut.register(test_widget1, 'TestWidget1',
                 tags={'userdef1': ['TestTag'], 'userdef2': ['TestTag']})
    dut.register(test_widget2, 'TestWidget2',
                 tags={'userdef1': ['TestTag'], 'userdef2': ['TestTag']})
    dut.register(test_widget3, 'TestWidget3',
                 tags={'userdef1': ['TestTag'], 'userdef2': ['TestTag2']})
    assert dut.lookup_by_tag_and(userdef1='TestTag', userdef2='TestTag') == {
        'testwidget1': test_widget1,
        'testwidget2': test_widget2,
    }


def test_lookup_multiple_tag_values() -> None:
    '''Find a widget that matches at least one tag value out of a list.'''
    dut = WidgetCatalog()
    test_widget1 = NamedWidget('TestWidget1')
    test_widget2 = NamedWidget('TestWidget2')
    test_widget3 = NamedWidget('TestWidget3')
    dut.register(test_widget1,
                 tags={'species': 'hamster'})
    dut.register(test_widget2,
                 tags={'species': ['gerbil', 'wolverine']})
    dut.register(test_widget3,
                 tags={'species': 'cuttlefish'})
    assert dut.lookup_by_tag_and(species=('gerbil', 'hamster')) == {
        'testwidget1': test_widget1,
        'testwidget2': test_widget2,
    }


def test_register_with_strings() -> None:
    '''Test registering a template with strings instead of lists in tags'''
    dut = WidgetCatalog()
    test_widget = NamedWidget('TestWidget')
    assert dut.register(test_widget, tags={'data_type': 'Some Type',
                        'task': 'Some Task'}) == 'testwidget'
    assert dut.lookup_by_tag_and(task='some task') == {'testwidget': test_widget}


def test_load() -> None:
    dut = WidgetCatalog(frob='hello', quux='world')
    dut.load(Path(__file__).parent / 'widgets')

    by_tag = dut.lookup_by_tag_and(some_tag='test_dir')
    assert set(by_tag.keys()) == set({'doodad', 'gizmo'})

    by_name = dut.lookup_by_name('doodad')
    assert hasattr(by_name, 'frob')
    assert hasattr(by_name, 'quux')
    assert by_name.frob == 'hello'
    assert by_name.quux == 'world'


class WidgetWithTags(NamedWidget):
    tags = {'some_tag': ['some_value']}


def test_tag_autoloading():
    dut = WidgetCatalog()
    widget = WidgetWithTags('widget_with_tags')
    dut.register(widget)

    by_tag = dut.lookup_by_tag_and(some_tag='some_value')
    assert set(by_tag.keys()) == set({'widget_with_tags'})


# Test registration with CatalogElementMixin. We're confirming
# that registration picks up _name and _tags.
class WidgetNamedTagged(Widget, CatalogElementMixin):
    _name = 'widget_named_tagged'
    _tags = {
        'tag_key': ['tag_value1', 'tag_value2'],
    }


def test_widget_named_tagged_succeeds():
    dut = WidgetCatalog()
    widget = WidgetNamedTagged()
    dut.register(widget)

    by_name = dut.lookup_by_name(name='widget_named_tagged')
    assert hasattr(by_name, 'name')
    assert by_name.name == widget.name

    by_tag = dut.lookup_by_tag_and(tag_key='tag_value2')
    assert by_tag[widget.name].name == widget.name


def test_lookup_items() -> None:
    '''Find a list of several widgets by arbitrary tag.'''
    dut = WidgetCatalog()
    test_widgetone = NamedWidget('TestWidget')
    dut.register(obj=test_widgetone, name='TestWidget')
    test_widgettwo = NamedWidget('TestWidgetTwo')
    dut.register(obj=test_widgettwo, name='TestWidgetTwo')

    assert dut.items() == {
        'testwidget': test_widgetone,
        'testwidgettwo': test_widgettwo,
    }.items()



'''Decider that measures elapsed time since last send to any neighbor.'''
# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from enum import Enum
import time
from typing import Mapping

from aenum import Enum as AEnum  # type: ignore[import-untyped]

from ..algorithms.impl.neighbor_metadata_interface import NeighborMetadataInterface
from ..config_components.distributed_config import DecidersConfig, DeciderConfig
from ..neighbor_manager.node_id import NodeID  # type: ignore[import-untyped]

from .impl.decider import Decider
from .impl.decider_catalog import DeciderCatalog


class TimeSinceSentDeciderConfig(DeciderConfig):
    '''Configuration for TimeSinceSentDecider.'''

    def __init__(self, name: str, clause: dict) -> None:
        super().__init__(name=name, clause=clause)

    def __str__(self) -> str:
        return f'time_seconds: {self.time_seconds}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TimeSinceSentDeciderConfig):
            return False
        return self.time_seconds == other.time_seconds

    class Defaults(Enum):
        '''Default values for the clause.'''
        TIME_SECONDS = 1.0

    class Keys(AEnum):
        '''Keys for the clause.'''
        ENABLED = 'enabled'
        TIME_SECONDS = 'time_seconds'

    @property
    def time_seconds(self) -> float:
        '''Get the refractory period.

        This is how long we wait before returning True.
        '''
        return self._get_with_default(
            self.Keys.TIME_SECONDS, dflt=self.Defaults.TIME_SECONDS.value)


class TimeSinceSentDecider(Decider):
    '''Decider that triggers a refractory time past the last sent message.'''
    name = 'time_since_sent'
    tags = {}
    _config: TimeSinceSentDeciderConfig

    def decide(self, my_meta: NeighborMetadataInterface,
               neighbor_id: NodeID, neighbors: Mapping[NodeID, NeighborMetadataInterface]) -> bool:
        '''Decide if we should send to neighbor.'''
        assert my_meta.current_state is not None
        interval = time.monotonic() - my_meta.last_state_sent_time
        return interval > self._config.time_seconds


def register(catalog: DeciderCatalog) -> None:
    '''Register all objects in this file.'''
    catalog.register(TimeSinceSentDecider)
    DecidersConfig.register(name=TimeSinceSentDecider.name, config_type=TimeSinceSentDeciderConfig)



'''Deal with hyperparam bindings.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from numbers import Number
from typing import Any, Callable, Dict, Iterable, List, Type

from ...wrangler.constants import RangeMethod


class Error(Exception):
    '''Base class for all errors thrown by this module'''


class RangeError(Error):
    '''Binder was passed inappropriate range of values.'''


class Binder(metaclass=abc.ABCMeta):
    '''Create a dict naming all hyperparameter values for a step.

    parse_value is an algorithm-specific function that parses JSON values of a
            parameter into python values.
    '''

    _parse_value: Callable[[str, Any], Any]

    def __init__(self, parse_value: Callable[[str, Any], Any]):
        self._parse_value = parse_value

    @abc.abstractmethod
    def bind(self, hyperparam_name: str, prange: Any) -> Dict[str, Any]:
        '''Create a dict naming all hyperparameter values for a step.

        The typing of prange depends on the value of method, and the
        type T (as parsed from JSON) of the hyperparameter
        associated with this type of Binder\\:

        * FIXED: T
        * LIST: List[T]
        * LINEAR: List[T]
            * 3 elements: [start, increment, end]
            * T must be a NumberLikeObject (supports +, < operations)
            * If any of these values are complex, we will use their
              absolute values for comparison.

        Keys of retval are hyperparam values parsed to str
        Values of retval are corresponding hyperparam values parsed into python
        '''


class FixedBinder(Binder):
    '''Bind FIXED ParamRange.'''

    def bind(self, hyperparam_name: str, prange: Any) -> Dict[str, Any]:
        '''Bind FIXED ParamRange.

        prange: Any
            single json hyperparam value

        Keys of retval are hyperparam values unparsed to str.
        Values of retval are corresponding hyperparam values parsed into python.

        retval will always have 1 element in it.
        '''
        return {str(prange): self._parse_value(hyperparam_name, prange)}


class ListBinder(Binder):
    '''Bind LIST ParamRange.'''

    def bind(self, hyperparam_name: str, prange: Any) -> Dict[str, Any]:
        '''Bind LIST ParamRange.

        prange: List[Any]
            List of json hyperparam values

        Keys of retval are hyperparam values unparsed to str.
        Values of retval are corresponding hyperparam values parsed into python.

        retval will contain a number of elements equal to the length of the list in
            the problem definition.
        '''
        if not isinstance(prange, Iterable):
            raise RangeError('ListBinder requires an iterable param range, '
                             f'instead found {prange} of type {type(prange)}.')
        retval: Dict[str, Any] = {}
        value_list = [
            self._parse_value(
                hyperparam_name,
                value) for value in prange]
        value_strings = [str(p) for p in prange]
        for i, value_str in enumerate(value_strings):
            retval[value_str] = value_list[i]
        return retval


def abs_if_complex(num: Number):
    '''Take abs of complex numbers only.'''
    if isinstance(num, complex):
        return abs(num)
    return num


class LinearBinder(Binder):
    '''Bind LINEAR ParamRange.'''

    def bind(self, hyperparam_name: str, prange: Any) -> Dict[str, Any]:
        '''Bind LINEAR ParamRange.

        prange: List[Any]
            a list of 3 values, [start, increment, max],
            which are all parsable into subclasses of Number.
            If any of these values are complex, we will use their absolute values for comparison.

        Keys of retval are hyperparam values unparsed to str.
        Values of retval are corresponding hyperparam values parsed into python.
        '''
        if not isinstance(prange, Iterable):
            raise RangeError('LinearBinder requires an iterable param range, '
                             f'instead found {prange} of type {type(prange)}.')
        prange = list(prange)
        if len(prange) != 3:
            raise RangeError('LinearBinder requires a param range of length 3, '
                             f'instead found {prange} of length {len(prange)}')
        retval: Dict[str, Any] = {}

        start, increment, end = [
            self._parse_value(
                hyperparam_name, value) for value in prange]
        values: List[Number] = []
        current = start
        # We conditionally use abs() so this works for complex numbers.
        while abs_if_complex(current) <= abs_if_complex(end):
            values.append(current)
            current += increment
        for value in values:
            retval[str(value)] = value
        # These assertions must come after the basic operations so that
        # mypy does not complain about Number not supporting +.
        if not isinstance(start, Number):
            raise RangeError('LinearBinder start must parse to a Number. '
                             f'Instead found {start} of type {type(start)}')
        if not isinstance(increment, Number):
            raise RangeError('LinearBinder increment must parse to a Number. '
                             f'Instead found {increment} of type {type(increment)}')
        if not isinstance(end, Number):
            raise RangeError('LinearBinder end must parse to a Number. '
                             f'Instead found {end} of type {type(start)}')
        return retval


BIND: Dict[RangeMethod, Type[Binder]] = {
    RangeMethod.FIXED: FixedBinder,
    RangeMethod.LINEAR: LinearBinder,
    RangeMethod.LIST: ListBinder,
}


class BinderFactory():
    '''Make a Binder instance for the given method.'''

    def build(self,
              range_method: RangeMethod,
              parse_value: Callable[[str, Any], Any]
              ) -> Binder:
        '''Build a Binder object for the given range_method.

        Pass in the Algorithm-specific value parser.
        '''
        if range_method not in BIND:
            raise NotImplementedError(
                f'BUG: method {range_method} not implemented.')

        return BIND[range_method](parse_value=parse_value)



'''Tests for data_loader_auto.py.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import os
from pathlib import Path

import pandas as pd
import pytest

from ...config_components.dataset_config import DatasetConfig
from ...wrangler.constants import PACKAGE_NAME
from ..memory_data_loader import MemoryDataLoader
from .data_loader import DataLoader
from .data_loader_auto import DataLoaderCatalogAuto

# pylint: disable=missing-function-docstring,redefined-outer-name,duplicate-code
# Note that we disable flake8 error F841 (# noqa F841)
# because we have a local variable which is used
# but which static analysis can't tell is used.


FAKE_DATA_LOADER_SOURCE = '''
from typing import Any, Dict, List, Optional
from ngautonml.wrangler.dataset import Dataset
from ngautonml.data_loaders.impl.data_loader import DataLoader
from ngautonml.data_loaders.impl.data_loader_catalog import DataLoaderCatalog

class FakeDataLoader(DataLoader):
    name = 'fake_data_loader'
    tags = {
        'input_format': ['some_input_format'],
        'loaded_format': ['some_loaded_format']
    }

    def _load_train(self) -> Dataset:
        return Dataset()

    def _load_test(self) -> Optional[Dataset]:
        return None

    def _load_ground_truth(self) -> Optional[Dataset]:
        return None

    def _dataset(self, data: Any, **kwargs) -> Dataset:
        return Dataset()

    def validate(self, dataset: Dataset) -> None:
        pass

    def poll(self, timeout: Optional[float] = 0.0) -> Optional[Dataset]:
        return Dataset()

    def build_dataset_from_json(self, json_data: Dict) -> Dataset:
        return Dataset()


def register(catalog: DataLoaderCatalog):
    catalog.register(FakeDataLoader)
'''


@pytest.fixture(scope="session")
def tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    retval = tmp_path_factory.mktemp("data")
    source_path = (retval / f'.{PACKAGE_NAME}' / 'plugins'
                   / 'data_loaders' / 'fake_data_loader.py')
    source_path.parent.mkdir(parents=True, exist_ok=True)
    with source_path.open('wt') as file:
        file.write(FAKE_DATA_LOADER_SOURCE)
    return retval


def test_data_loader_catalog_auto() -> None:
    config = DatasetConfig(clause={
        'params': {
            'train_data': 'train_df'
        }
    })
    train_df = pd.DataFrame()  # noqa: F841 pylint: disable=unused-variable
    dut = DataLoaderCatalogAuto()

    got = dut.lookup_by_name('memory')(config=config)
    assert isinstance(got, MemoryDataLoader)


def test_home_directory_auto(tmp_path) -> None:
    os.environ['HOME'] = str(tmp_path)
    config = DatasetConfig(clause={
        'params': {
            'train_data': 'train_df'
        }
    })
    train_df = pd.DataFrame()  # noqa: F841 pylint: disable=unused-variable

    dut = DataLoaderCatalogAuto()
    got = dut.lookup_by_name('fake_data_loader')(config=config)
    assert isinstance(got, DataLoader)
    assert got.tags['input_format'] == ['some_input_format']


def test_lookup_by_formats() -> None:
    config = DatasetConfig(clause={
        'params': {
            'train_data': 'train_df'
        }
    })
    train_df = pd.DataFrame()  # noqa: F841 pylint: disable=unused-variable

    dut = DataLoaderCatalogAuto()
    got = dut.lookup_by_formats(input_format='pandas_dataframe', loaded_format='pandas_dataframe')
    assert any(isinstance(cons(config=config), MemoryDataLoader) for cons in got.values())


def test_construct_instance() -> None:
    config = DatasetConfig(clause={
        'config': 'memory',
        'params': {
            'train_data': 'train_df'
        }
    })
    train_df = pd.DataFrame()  # noqa: F841 pylint: disable=unused-variable

    dut = DataLoaderCatalogAuto()
    got = dut.construct_instance(config=config)
    assert isinstance(got, MemoryDataLoader)



'''Loads an in-memory pandas DataFrame'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Dict, Optional, List

import pandas as pd

from ..wrangler.dataset import Dataset
from ..config_components.dataset_config import DatasetConfig
from ..config_components.impl.config_component import (
    InvalidValueError, MissingKeyError)
from .impl.dataframe_loader import DataframeLoader
from .impl.data_loader_catalog import DataLoaderCatalog


class MemoryDataLoader(DataframeLoader):
    '''Loads an in-memory pandas DataFrame'''

    name: str = 'memory'
    tags: Dict[str, List[str]] = {
        'input_format': ['pandas_dataframe'],
        'loaded_format': ['pandas_dataframe'],
    }

    def __init__(self, config: DatasetConfig):
        if DatasetConfig.Keys.TRAIN_DATA.value not in config.params:  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            # TODO(Merritt/Piggy): add test for this error state
            raise MissingKeyError(
                'Attempt to create a memory dataloader, but train_data is not provided.'
            )

        train_df_var_name = config.params[DatasetConfig.Keys.TRAIN_DATA.value]   # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
        self._train_df = self._lookup_var(train_df_var_name)
        if not isinstance(self._train_df, pd.DataFrame):
            # TODO(Merritt/Piggy): add test for this error state
            raise InvalidValueError(
                f'Expecting local variable {train_df_var_name} to be '
                f'a pandas.DataFrame; instead found {type(self._train_df)}'
            )

        self._test_df = None
        if DatasetConfig.Keys.TEST_DATA.value in config.params:  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            test_df_var_name = config.params[DatasetConfig.Keys.TEST_DATA.value]  # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            self._test_df = self._lookup_var(test_df_var_name)
            if not isinstance(self._test_df, pd.DataFrame):
                # TODO(Merritt/Piggy): add test for this error state
                raise InvalidValueError(
                    f'Expecting local variable {test_df_var_name} to be '
                    f'a pandas.DataFrame; instead found {type(self._test_df)}'
                )

        super().__init__(config=config)

    def _load_train(self) -> Dataset:
        retval = Dataset(metadata=self._metadata)
        retval.dataframe = self._train_df
        # TODO(Piggy/Merritt): this is a hack for forecasting,
        # replace with something that can be used by both memory and local
        # to load additional data requested by a plugin
        retval['static_exogenous'] = None

        return retval

    def _load_test(self) -> Optional[Dataset]:
        if self._test_df is None:
            return None

        retval = Dataset(metadata=self._metadata)
        retval.dataframe = self._test_df

        # TODO(Piggy/Merritt): this is a hack for forecasting,
        # replace with something that can be used by both memory and local
        # to load additional data requested by a plugin
        retval['static_exogenous'] = None

        return retval


def register(catalog: DataLoaderCatalog):
    '''Register all the objects in this file.'''
    catalog.register(MemoryDataLoader)



'''Metric object class for configurable and user-made metrics.'''
import abc

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from typing import Optional

from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...wrangler.dataset import Dataset
from ...wrangler.dataset import DatasetKeys


class MetricError(Exception):
    '''Base error class for metrics'''


class MetricNameError(MetricError):
    '''Metric lacks a proper indexing name.'''


class MetricAttributeError(MetricError):
    '''Metric is missing a required attribute.'''


class MetricInvalidDatasetError(MetricError):
    '''Metric was passed a dataset it cannot use.'''


class Metric(CatalogElementMixin, metaclass=abc.ABCMeta):
    '''Base class wrapper for accessing metric objects.

    **Standard tags\\:**

    high
        values in 'true', 'false'.
        'true' means a higher score is better.

        If not specified, 'high' is assumed to be 'true' by the Ranker.

    needs_pos_label
        values in 'true', 'false'.
        'true' means there must be a pos_label specified in the Metadata
        in order for the metric to be used.

    needs_proba
        values in 'true', 'false'

        True means metric uses class probabilites instead of class predictions.
        For ex., ROC AUC.
        Should only be True for classification metrics.

    task
        values in 'binary_classification', 'multiclass_classification',
        'regression', and any other existing task types

        task(s) that the metric can be used for.
    '''
    _name: Optional[str] = None

    def __str__(self):
        return self.name

    @property
    def high(self) -> bool:
        '''Interprets 'high' tag as a boolean.

        True means a higher score is better.
        Defaults to True if tag is not set.
        '''
        return self._tag_to_bool('high', default=True)

    @property
    def needs_pos_label(self) -> bool:
        '''Interprets the 'needs_pos_label' tag as a boolean.

        False means there must be a pos_label specified in the Metadata
        in order for the metric to be used.

        Defaults to True if tag not set.
        '''
        return self._tag_to_bool('needs_pos_label', default=False)

    @property
    def needs_proba(self) -> bool:
        '''Interprets needs_proba tag as a boolean.

        True means metric uses class probabilites instead of class predictions.
        Defaults to False if tag not set
        '''
        return self._tag_to_bool('needs_proba', default=False)

    @abc.abstractmethod
    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        '''Calculates the metric results.'''

    def calculate_roc_curve(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> Dataset:
        '''Calculate the ROC Curve.'''
        _, _ = pred, ground_truth  # Convince pylint that this function is not abstract.
        raise NotImplementedError(
            f'calculate_roc_curve is not implemented for metric {self.name}.')


class SupervisedMetric(Metric):
    '''Subclass for supervised Metrics'''

    def validate_input(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> Dataset:
        '''Validates the Dataset and GroundTruth given. Useful for derived classes.'''
        if ground_truth is None:
            raise MetricInvalidDatasetError(
                'No Ground Truth Provided!')

        if pred is None:
            raise MetricInvalidDatasetError(
                'No Predictions Provided!')

        if DatasetKeys.PREDICTIONS.value not in pred:
            raise MetricInvalidDatasetError(
                f'Predictions has no "{DatasetKeys.PREDICTIONS.value}" key.')

        # Attempt to infer the GROUND_TRUTH.
        # TODO(Merritt/Piggy): Move this functionality to Dataset.
        if DatasetKeys.GROUND_TRUTH.value not in ground_truth:
            out = ground_truth.output()
            if DatasetKeys.TARGET.value in ground_truth:
                out[DatasetKeys.GROUND_TRUTH.value] = ground_truth[DatasetKeys.TARGET.value]
            elif DatasetKeys.PREDICTIONS.value in ground_truth:
                out[DatasetKeys.GROUND_TRUTH.value] = ground_truth[DatasetKeys.PREDICTIONS.value]
            elif DatasetKeys.DATAFRAME.value in ground_truth:
                if ground_truth.metadata.target is None:
                    raise MetricInvalidDatasetError(
                        'Ground truth has no target column in metadata for a dataframe.')
                out[DatasetKeys.GROUND_TRUTH.value] = ground_truth[
                    DatasetKeys.DATAFRAME.value][ground_truth.metadata.target.name]
            else:
                raise MetricInvalidDatasetError(
                    f'Ground truth has no "{DatasetKeys.GROUND_TRUTH.value}" key.')
            ground_truth = out

        if ground_truth.metadata.target is None:
            raise MetricInvalidDatasetError(
                'Ground truth has no target column in metadata.')

        # We take a Optional[Dataset] and turn it into a Dataset. Type checkers are happy.
        return ground_truth

    @abc.abstractmethod
    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        '''
        Calculates the metric results.

        Most Supervised Metrics will require a ground truth. The implementation
        of the calculate should check if the ground truth exists.
        '''


class MetricStub(Metric):
    '''This is a stub'''
    _name = 'stub_metric'
    _high = True

    def calculate(self, pred: Dataset, ground_truth: Optional[Dataset] = None) -> float:
        return 0.0



'''Base class for sklearn models'''
# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=duplicate-code,too-many-arguments
from typing import Any, Dict, List, Optional
import importlib

import numpy as np
from numpy.random import RandomState  # pylint: disable=no-name-in-module  # This just isn't true.
import pandas as pd
from sklearn.base import BaseEstimator  # type: ignore[import-untyped]

from ...problem_def.task import DataType, Task, TaskType
from ...tables.impl.table import TableFactory
from ...wrangler.constants import Defaults
from ...wrangler.dataset import Dataset, DatasetKeys

from ..impl.algorithm import Algorithm, AlgorithmCatalog
from ..impl.sklearn_like_algorithm_instance import SklearnLikeAlgorithmInstance
from ..impl.algorithm_instance import DatasetError
from ..impl.fittable_algorithm_instance import UntrainedError


def is_empty(frame) -> bool:
    '''Determine if an array-like object is empty.'''
    if frame is None:
        return True
    if isinstance(frame, np.ndarray) and frame.size == 0:
        # ndarray.size represents the number of elements in the array
        # ndarray.ndim represents the number of dimensions in the array
        #   arrays with more than 0 dimension can have 0 elements, and
        #   we consider those empty
        return True
    if isinstance(frame, pd.DataFrame) and frame.empty:
        return True
    return False


# TODO(Merritt): set random seed
class SklearnAlgorithm(Algorithm):
    '''Generic object for an sklearn algorithm.'''
    def __init__(self,
                 name: str,
                 algorithm: Optional[type] = None,
                 tasks: Optional[List[TaskType]] = None,
                 data_types: Optional[List[DataType]] = None,
                 tags: Optional[Dict[str, List[str]]] = None,
                 **hyperparams: Any):
        self._name = name
        if algorithm is None:
            algorithm = self._load_module(name)
        self._algorithm = algorithm
        if tasks is None:
            tasks = []
        if data_types is None:
            data_types = []
        if tags is not None:
            self._tags = tags.copy()
        else:
            self._tags = {}
        self._tags.update({
            Task.Keys.TASK_TYPE.value: [t.name.lower() for t in tasks],   # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            Task.Keys.DATA_TYPE.value: [d.name.lower() for d in data_types],   # type: ignore[attr-defined] # pylint: disable=no-member,line-too-long
            'source': ['sklearn'],
        })
        super().__init__(**hyperparams)

    def _load_module(self, name: str):
        # Split name into module part (e.g. sklearn.linear_model)
        # and constructor part (e.g. LinearRegression)
        parts = name.split('.')
        constructor_part = parts[-1]
        module = importlib.import_module('.'.join(parts[:-1]))
        # Load the constructor.
        return getattr(module, constructor_part)

    def instantiate(self, **hyperparams) -> 'SklearnAlgorithmInstance':
        # TODO(Merritt): test that we're actually passing hyperparams through
        return SklearnAlgorithmInstance(
            parent=self,
            constructor=self._algorithm,
            **self._default_param_ranges(**self.hyperparams(**hyperparams)))


class SklearnAlgorithmInstance(SklearnLikeAlgorithmInstance):
    '''Generic object for an sklearn algorithm.'''

    def __init__(self, parent: Algorithm, constructor: type, *args, **kwargs):
        self._constructor = constructor
        assert parent._tags is not None, (
            'BUG: Algorithm __init__ should have set _tags'
        )
        if parent._tags.get('supports_random_seed', ['false'])[0] == 'true':
            if 'random_seed' in kwargs:
                kwargs['random_state'] = RandomState(kwargs.pop('random_seed'))
            elif kwargs.get('random_state', None) is None:
                kwargs['random_state'] = RandomState(Defaults.SEED)
        super().__init__(parent, *args, **kwargs)

    def fit(self, dataset: Optional[Dataset]):
        if dataset is None:
            raise DatasetError(f'attempt to fit with no data for {self.catalog_name}')

        # TODO(piggy): This logic belongs in Dataset.target and Dataset.covariates, not here.
        # This code allows a single 'dataframe' table instead of separate 'target' and 'covariates'.
        if DatasetKeys.TARGET.value not in dataset:
            if dataset.metadata.target is not None:
                target_for_sklearn = dataset.dataframe[dataset.metadata.target.name]
            else:
                # This is for fitting algorithms that don't need a target.
                target_for_sklearn = None
        else:
            target_data = dataset[DatasetKeys.TARGET.value]
            if isinstance(target_data, pd.DataFrame):
                assert dataset.metadata.target is not None
                target_for_sklearn = target_data[dataset.metadata.target.name]
            else:
                target_for_sklearn = pd.Series(list(target_data))

        if DatasetKeys.COVARIATES.value not in dataset:
            covariate_data = dataset.dataframe
            if isinstance(covariate_data, pd.DataFrame):
                assert dataset.metadata.target is not None
                covariates = dataset.dataframe.drop(
                    dataset.metadata.target.name, axis=1)
            else:
                covariates = list(covariate_data)
        else:
            covariates = dataset[DatasetKeys.COVARIATES.value]

        new_dataset = dataset.output()
        new_dataset[DatasetKeys.TARGET.value] = target_for_sklearn
        new_dataset[DatasetKeys.COVARIATES.value] = covariates
        super().fit(dataset=new_dataset)

    def predict(self, dataset: Optional[Dataset]) -> Optional[Dataset]:
        if not self._trained:
            raise UntrainedError(f'attempt to predict before fit for {self.catalog_name}')

        if dataset is None:
            return None

        # this is not typed because it is supposed to hold the return value from sklearn
        # we expect it to be np.ndarray, but there is no guarantee in sklearn
        predictions_result = np.empty(0)
        probabilities_result = np.empty((0, 0))

        try:
            df = None
            if DatasetKeys.COVARIATES.value in dataset:
                df = dataset[DatasetKeys.COVARIATES.value]
            elif DatasetKeys.DATAFRAME.value in dataset:
                df = dataset[DatasetKeys.DATAFRAME.value]
                if (dataset.metadata.target is not None
                        and dataset.metadata.target.name in df.columns):
                    df.drop(dataset.metadata.target.name, axis=1, inplace=True)

            if not is_empty(df):
                predictions_result = self._impl.predict(df)

                # assume the algorithm has a predict_proba function unless otherwise marked
                has_predict_proba = 'false' not in self.algorithm.tags.get(
                    'has_predict_proba', ['true'])
                if dataset.metadata.task == TaskType.BINARY_CLASSIFICATION and has_predict_proba:
                    probabilities_result = self._impl.predict_proba(
                        dataset[DatasetKeys.COVARIATES.value])

        except KeyError as err:
            raise DatasetError(f'predict dataset malformed {dataset!r}') from err

        # sklearn does not use typed Python, and thus we do not trust
        #    its return value to be consistently typed
        if is_empty(predictions_result):
            predictions_result = np.empty(0)

        retval = dataset.output()
        assert dataset.metadata.target is not None, (
            'BUG: predict() called with no target info in metadata'
        )
        retval.predictions = pd.DataFrame({
            dataset.metadata.target.name: predictions_result
        })

        if not is_empty(probabilities_result):
            # we arbitrarily choose only one column of probabilities to be included in retval
            probabilities_result = np.delete(probabilities_result, 0, axis=1).flatten()  # type: ignore[assignment] # pylint: disable=line-too-long
            retval.probabilities = TableFactory({
                dataset.metadata.target.name: probabilities_result
            })

        return retval

    @property
    def impl(self) -> BaseEstimator:
        '''Get the underlying sklearn model.'''
        assert isinstance(self._impl, BaseEstimator)
        return self._impl


def register(catalog: AlgorithmCatalog, **kwargs):  # pylint: disable=unused-argument
    '''Nothing to register.

    All subclasses of SklearnAlgorithm are registered in sklearn_algorithms.py
    '''



'''Autoloader for the PipelineLoader catalog.'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

from pathlib import Path
from typing import Optional

from ...catalog.plugin_catalog import PluginCatalog
from .pipeline_loader import PipelineLoader
from .pipeline_loader_catalog import PipelineLoaderCatalog


class PipelineLoaderCatalogAuto(PluginCatalog[PipelineLoader], PipelineLoaderCatalog):
    '''PipelineLoader catalog that loads from $HOME and from the installation directory.'''
    def __init__(self, default_root: Optional[Path] = None, **kwargs):
        if default_root is None:
            default_root = Path(__file__).parents[2]
        super().__init__(catalog_name='pipeline_loaders', default_root=default_root, **kwargs)



'''Distributed Linear Regression'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# Linear and Logistic Regression have portions of identical code that should
# not be abstracted away.
# pylint: disable=duplicate-code

import pickle
from typing import Callable, Iterable, Optional, Tuple

import numpy as np


from ...catalog.catalog import upcast
from ...problem_def.task import DataType, TaskType
from ...wrangler.dataset import Dataset, DatasetKeys, RoleName
from ...wrangler.logger import Logger

from ..impl.algorithm import Algorithm, MemoryAlgorithmCatalog
from ..impl.distributed_algorithm_instance import NoDataError
from ..impl.linear_distributed_algorithm_instance import (
    LinearDistributedAlgorithmInstance, LinearNeighborState,
    norm2, inner_product)


logger = Logger(__file__).logger()


class AutonLinearRegressionNeighbor(LinearNeighborState):
    '''A neighbor in the distributed linear regression model.'''

    def encode(self) -> bytes:
        '''Encode message for distributed neighbors.'''
        #  minimal: communicate as little information as possible
        return pickle.dumps((self._v, self.columns))

    @classmethod
    def decode(cls, serialized_model: bytes) -> 'AutonLinearRegressionNeighbor':
        '''Decode a message from a neighbor.'''
        got = pickle.loads(serialized_model)
        assert len(got) == 2, (
            f'BUG: expected to recv a tuple of length 2, instead found {got} of length {len(got)}.'
        )
        (v, cols) = got
        assert isinstance(v, np.ndarray), (
            f'BUG: expected v to be an np.ndarray, instead found {v} of type {type(v)}'
        )
        if cols is not None:
            cols = cls._cast_columns(columns=cols)
        return cls(v=v, columns=cols)


class AutonLinearRegressionInstance(LinearDistributedAlgorithmInstance):
    '''Binary logistic regression that supports distributed AI.

    As a distributed model, this instance can share information about its
    trained state with other instances and update its trained state using
    information shared by other instances.
    '''
    _neighbor_constructor = AutonLinearRegressionNeighbor

    # This is a type annotation for _my_state.
    _predict_state: Optional[AutonLinearRegressionNeighbor] = None

    @property  # type: ignore[override]
    def _my_state(self) -> Optional[AutonLinearRegressionNeighbor]:
        retval = LinearDistributedAlgorithmInstance._my_state.fget(self)  # type: ignore[attr-defined] # pylint: disable=assignment-from-no-return,line-too-long
        assert retval is None or isinstance(retval, AutonLinearRegressionNeighbor), (
            'BUG: expected _my_state to be None or an AutonLinearRegressionNeighbor.')
        return retval

    @_my_state.setter
    def _my_state(self, value: Optional[AutonLinearRegressionNeighbor]) -> None:
        assert value is None or isinstance(value, AutonLinearRegressionNeighbor), (
            'BUG: expected value to be None or an AutonLinearRegressionNeighbor.')
        LinearDistributedAlgorithmInstance._my_state.fset(self, value)  # type: ignore[attr-defined]

    @property
    def _yhat_is_proba(self) -> bool:
        return False

    @property
    def _neighbor_models_iter(self) -> Iterable[AutonLinearRegressionNeighbor]:
        for v in super()._neighbor_models_iter:
            assert isinstance(v, AutonLinearRegressionNeighbor), (
                'BUG: expected neighbor_models to contain AutonLinearRegressionNeighbor, '
                f'instead found {v} of type {type(v)}'
            )
            yield v

    def _standardize_y(self, y: np.ndarray) -> np.ndarray:
        '''Regularize the target values.'''
        return y

    def _yhat(self, x: np.ndarray) -> np.ndarray:
        assert self._predict_state is not None, (
            'BUG: self._predict_state should not be None when _yhat is called.'
            ' This should have been set in _fit'
        )
        return self._predict_state.v @ x.T

    def _decode(self, serialized_model: bytes) -> AutonLinearRegressionNeighbor:
        '''Decode a message from distributed neighbors.'''
        return AutonLinearRegressionNeighbor.decode(serialized_model)

    def _prepare_data(self, dataset: Optional[Dataset]) -> Tuple[np.ndarray, np.ndarray]:
        '''Extract covariates and target, accounting for the possibility that there's no data.'''
        if dataset is None:
            if list(self._neighbor_models_iter):
                d = max(len(m.v) for m in self._neighbor_models_iter) - 1
                np_cov = np.zeros((0, d))
            else:
                # TODO(Piggy/Dan) Need to figure out how to deal with no neighbors and no data.
                raise NoDataError("No data and no neighbors.")
            np_tar = np.zeros(0)
        else:
            target_col_name = dataset.metadata.roles[RoleName.TARGET][0].name

            if DatasetKeys.DATAFRAME.value in dataset:
                orig_cov = dataset[DatasetKeys.DATAFRAME.value].copy()
                if target_col_name in orig_cov.columns:
                    orig_cov.drop(columns=[target_col_name], inplace=True)
                    orig_tar = dataset[DatasetKeys.DATAFRAME.value].loc[:, target_col_name]
                    np_tar = orig_tar.to_numpy()
                else:
                    np_tar = np.array([0])  # type: ignore[assignment]
            else:
                orig_cov = dataset[DatasetKeys.COVARIATES.value]
                if DatasetKeys.TARGET.value in dataset:
                    orig_tar = dataset[DatasetKeys.TARGET.value]
                    np_tar = orig_tar[target_col_name].to_numpy()
                else:
                    np_tar = np.array([0])  # type: ignore[assignment]

            np_cov = orig_cov.to_numpy()

        # Add a column for the intercept.
        np_cov = np.hstack((np_cov, np.ones((np_cov.shape[0], 1))))  # type: ignore[assignment]

        _, d = np_cov.shape

        if self._my_state is None:
            self._my_state = AutonLinearRegressionNeighbor(
                v=np.zeros(d),
                columns=self._columns)

        return np_cov, np_tar

    def _objective_with_gradient(
            self, x: np.ndarray, y: np.ndarray, v_old: np.ndarray,
            l2: np.float64, _lambda: np.float64, omega: np.float64,
            k: int) -> Tuple[Callable[[np.ndarray], np.float64],
                             Callable[[np.ndarray], np.ndarray]]:
        '''Return the objective function and:190
          its gradient.'''

        if k > 0:
            def fun(v: np.ndarray) -> np.float64:
                data_loss = (
                    np.square(y - ((v @ x.T))).sum()
                )
                l2_regularization = l2 * norm2(v[:-1])
                # TODO(Merritt): Jack doesn't think dividing by k works here since
                #   neighbors will weight each other non-symmetrically.
                # Send information about neighbors' neighbors?
                neighbor_regularization = (
                    -2 * _lambda / k
                    * sum(inner_product(m.v, v) for m in self._neighbor_models_iter)
                )
                self_regularization = (
                    -2 * _lambda * (1 - omega) / omega * inner_product(v_old, v)
                )
                shared_regularization = (_lambda / omega) * norm2(v)
                return (data_loss
                        + l2_regularization
                        + neighbor_regularization
                        + self_regularization
                        + shared_regularization)

            def jac(v: np.ndarray) -> np.ndarray:
                data_loss_grad = (2 * (v @ x.T - y) * x.T).sum(1)
                # TODO(Merritt): maybe we can do this without allocating an extra array
                v_zero_intercept = v.copy()
                v_zero_intercept[-1] = 0
                l2_regularization_grad = 2 * l2 * v_zero_intercept
                neighbor_regularization_grad = -2 * _lambda / k * sum(
                    m.v for m in self._neighbor_models_iter)
                self_regularization_grad = -2 * _lambda * (1 - omega) / omega * v_old
                shared_regularization_grad = 2 * (_lambda / omega) * v
                return (data_loss_grad
                        + l2_regularization_grad
                        + neighbor_regularization_grad
                        + self_regularization_grad
                        + shared_regularization_grad)

        else:
            def fun(v: np.ndarray) -> np.float64:
                data_loss = np.square(y - (v @ x.T)).sum()
                l2_regularization = l2 * norm2(v[:-1])
                return data_loss + l2_regularization

            def jac(v: np.ndarray) -> np.ndarray:
                data_loss_grad = (2 * (v @ x.T - y) * x.T).sum(1)
                v_zero_intercept = v.copy()
                v_zero_intercept[-1] = 0
                l2_regularization_grad = 2 * l2 * v_zero_intercept
                return (data_loss_grad
                        + l2_regularization_grad)

        return fun, jac


class AutonLinearRegression(Algorithm):
    '''Class for Auton Lab's implementation of Linear Regression'''
    _name = "auton_linear_regression"
    _tags = {
        'tasks': [TaskType.REGRESSION.name],
        'data_types': [DataType.TABULAR.name],
        'source': ['auton_lab'],
        'distributed': ['true']
    }
    _instance_constructor = AutonLinearRegressionInstance
    _default_hyperparams = {
        'Lambda': 100.0,  # Neighbor regularization
        'L2': 1.0,  # The normal weight regularization
        'omega': 2.0 / 3.0,  # Self regularization with last iteration's local weights
        'tol': 0.000000001,
        'maxiter': None,
    }

    def instantiate(self, **hyperparams) -> 'AutonLinearRegressionInstance':
        return super().instantiate(**hyperparams)


def register(catalog: MemoryAlgorithmCatalog, *args, **kwargs) -> None:
    '''Register all the objects in this file.'''
    model = AutonLinearRegression(*args, **kwargs)
    catalog.register(model, model.name, upcast(model.tags))



'''Base class for objects that split datasets.'''

# Copyright (c) 2023 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import abc
from typing import Any, Dict, List, NamedTuple, Optional

import pandas as pd


from ...catalog.memory_catalog import MemoryCatalog
from ...catalog.catalog_element_mixin import CatalogElementMixin
from ...problem_def.cross_validation_config import CrossValidationConfig
from ...wrangler.dataset import Dataset, DatasetKeys


class SplitterError(Exception):
    '''Base error for splitters.'''


class Fold(NamedTuple):
    '''A fold for cross-validation'''
    train: Dataset
    validate: Dataset
    ground_truth: Optional[Dataset] = None


class SplitDataset(List[Fold]):
    '''A split dataset for cross-validation'''

    @property
    def folds(self) -> List[Fold]:
        '''List of folds

        Exists as a property for backwards compatibility.
        '''
        return self

    @property
    def ground_truth(self) -> Optional[Dataset]:
        '''Return ground truth if it exists and can be automatically extracted.'''
        if len(self) == 0:
            return None
        reference_dataset = self[0].train
        if DatasetKeys.DATAFRAME.value not in reference_dataset.keys():
            # Not a pandas dataset.
            return None

        if all(f.ground_truth is None for f in self):
            # Unsupervised
            return None

        dfs: List[pd.DataFrame] = []
        for f in self:
            assert f.ground_truth is not None
            dfs.append(f.ground_truth.ground_truth)

        retval = reference_dataset.output()
        retval.ground_truth = pd.concat(dfs).reset_index(drop=True)
        return retval


class Splitter(CatalogElementMixin, metaclass=abc.ABCMeta):
    '''Base class for objects that split datasets.'''

    _hyperparams: Dict[str, Any]
    _cv_config: CrossValidationConfig

    def __init__(self,
                 *args,
                 cv_config: CrossValidationConfig,
                 **default_hyperparams):
        super().__init__(*args, **default_hyperparams)
        self._cv_config = cv_config
        if not hasattr(self, '_hyperparams'):
            self._hyperparams = {}
        self._hyperparams.update(**default_hyperparams)

    def hyperparams(self, **overrides) -> Dict[str, Any]:
        '''Return splitter hyperparams, overridden with overrides'''
        retval = self._hyperparams.copy()
        retval.update(**overrides)
        return retval

    @abc.abstractmethod
    def split(self,
              dataset: Dataset,
              **overrides) -> SplitDataset:
        '''Divide dataset into multiple datasets according to the DatasetConfig.'''


class SplitterCatalog(MemoryCatalog[Splitter], metaclass=abc.ABCMeta):
    '''Base class for splitter catalog'''

    def lookup_by_task(self, task: str) -> Splitter:
        '''Get the right splitter for the task.'''
        splitters = self.lookup_by_tag_and(task=task)
        assert len(splitters) <= 1, f'BUG: Somehow we got too many splitters for task "{task}".'
        if len(splitters) == 0:
            splitters = self.lookup_by_tag_and(default='true')
            assert len(splitters) == 1, (
                'BUG: There should be exactly one default splitter (task "{task}").')
        return list(splitters.values())[0]



'''Tests for auton_decision_tree.py'''

# Copyright (c) 2024 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

# pylint: disable=missing-function-docstring,duplicate-code,too-many-locals,too-many-lines

from copy import deepcopy
import getpass
import os
from pathlib import Path
import pickle
import socket
from typing import Any, Dict, Optional, Tuple, List

import numpy as np
import pandas as pd
import pytest

from sklearn import datasets  # type: ignore[import]

from ...algorithms.impl.synchronous import advance
from ...config_components.distributed_config import DistributedConfig
from ...conftest import Clearer, Waiter
from ...metrics.impl.metric_auto import MetricCatalogAuto
from ...metrics.impl.metric import Metric
from ...problem_def.task import TaskType
from ...tables.impl.table_auto import TableCatalogAuto
from ...wrangler.base_port import BasePort
from ...wrangler.constants import Defaults
from ...wrangler.dataset import Column, Dataset, Metadata, RoleName
from ...wrangler.logger import Logger
from ..impl.algorithm_auto import AlgorithmCatalogAuto
from ..sklearn.sklearn_algorithm import SklearnAlgorithm
from .auton_decision_tree import AutonDecisionTreeModel, AutonDecisionTreeInstance
from .auton_decision_tree_neighbor import Acorn, AutonDecisionTreeNeighbor

logger = Logger(__file__, to_stdout=False).logger()
_ = TableCatalogAuto()  # pylint: disable=pointless-statement

BOOTSTRAP_SERVERS = os.environ.get('BOOTSTRAP_SERVERS', 'localhost:9092')


def topic_name(topic: str) -> str:
    '''Generate a unique topic name for the given test.'''
    return f'{Path(__file__).stem}_{getpass.getuser()}_{topic}'


DT_PARAMS = {
    'Lambda': 1.0,
    'omega': 0.667,
}


def load_classification_dataset() -> Tuple[Dataset, Dataset, Dataset, Dataset]:
    # TODO(Merritt): pull this out into a fixture
    # Load the breast cancer dataset
    bc_x_full, bc_y_series = datasets.load_breast_cancer(return_X_y=True, as_frame=True)
    assert isinstance(bc_x_full, pd.DataFrame)
    assert isinstance(bc_y_series, pd.Series)
    bc_y = pd.DataFrame({'target': bc_y_series})

    # restrict number of attributes for wider variability of results
    bc_x = bc_x_full.iloc[:, :3]

    test_size = 50
    # Split the data into training/testing sets
    bc_x_train = bc_x[:-test_size]
    bc_x_test = bc_x[-test_size:]

    # Split the targets into training/testing sets
    bc_y_train = bc_y[:-test_size]
    bc_y_test = bc_y[-test_size:].reset_index()

    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.BINARY_CLASSIFICATION
    )

    dataset_train = Dataset(
        metadata=metadata,
        covariates=bc_x_train,
        target=bc_y_train
    )

    dataset_test = Dataset(
        metadata=metadata,
        covariates=bc_x_test
    )
    ground_truth = Dataset(
        metadata=metadata,
        ground_truth=bc_y_test
    )

    train_cov_0 = dataset_train['covariates'][np.array(dataset_train['target']) == 0]
    train_cov_1 = dataset_train['covariates'][np.array(dataset_train['target']) == 1]
    reduced_train = Dataset(
        covariates=pd.concat((train_cov_0.iloc[0:1], train_cov_1.iloc[0:1]), axis=0) * 0,
        target=pd.DataFrame({'target': [0] * 1 + [1] * 1}),
        metadata=metadata
    )

    return (dataset_train, dataset_test, ground_truth, reduced_train)


def build_clause(adjacency: Dict[str, List[int]],
                 nodes_and_endpoints: List[Tuple[str, Tuple[str, int]]],
                 my_id: int,
                 data_split: Optional[Dict[str, int]] = None,) -> Dict[str, Any]:
    retval = {
        'polling_interval': '0.1',
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': adjacency
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': nodes_and_endpoints
            }
        },
        'my_id': my_id
    }
    if data_split:
        retval['split'] = data_split
    return retval


base_port = BasePort()
LOOPBACK = '127.0.0.1'
VALID_ACORN = Acorn(
    n_in=4,
    n_out=2,
    children_left=[1, None, None],
    children_right=[2, None, None],
    feature=[0, None, None],
    threshold=[0.5, None, None],
    equal=[True, None, None],
    value=[np.array([0.5, 0.5]), np.array([1., 0]), np.array([0, 1])],
    sample_weight=[13.0, 10.0, 3.0],
    lower_bound=np.array([0., 0., 0., 0.]),
    upper_bound=np.array([1., 1., 1., 1.]),
    class_list=['no', 'yes']
)


def test_should_send() -> None:
    '''_should_send returns True iff our model changed.

    We send a simple tree with 3 nodes.
    '''

    config = DistributedConfig({})
    dut = AutonDecisionTreeNeighbor(VALID_ACORN)
    same_as_dut = deepcopy(dut)
    different_from_dut = deepcopy(dut)
    different_from_dut.feature[0] = 3  # split on a different feature

    assert dut.state_differs(
        distributed=config,
        other=different_from_dut
    ) is True, 'Different states should differ.'

    assert dut.state_differs(
        distributed=config,
        other=same_as_dut
    ) is False, 'Same states should not differ.'


COLS_SENDER_PORT = base_port.next()
COLS_RECEIVER_PORT = base_port.next()


def test_columns(assert_no_exceptions: Clearer) -> None:
    '''Test that columns are handled in data loading, sending and receiving.'''
    distributed_config = DistributedConfig(clause=build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[
            ('1', (LOOPBACK, COLS_SENDER_PORT)),
            ('2', (LOOPBACK, COLS_RECEIVER_PORT)),
        ],
        my_id=2
    ))

    train = Dataset(
        covariates=pd.DataFrame({
            'a': [1.0, 2.0, 3.0],
            'b': [4.0, 5.0, 6.0],
            'c': [7.0, 8.0, 10.0],
            'd': [23.0, 63.0, 13.0]
        }),
        target=pd.DataFrame({
            'e': ['yes', 'no', 'no']
        }),
        metadata=Metadata(
            roles={RoleName.TARGET: [Column('e')]},
            task=TaskType.BINARY_CLASSIFICATION
        ))
    dut = AutonDecisionTreeModel().instantiate(distributed=distributed_config,
                                               synchronous=True,
                                               **DT_PARAMS)

    sock = socket.socket(socket.AF_INET, type=socket.SOCK_DGRAM)
    sock.bind((LOOPBACK, COLS_SENDER_PORT))

    match_message = pickle.dumps((
        VALID_ACORN,
        train['covariates'].columns.to_list()
    ))

    mismatch_message = pickle.dumps((
        VALID_ACORN,
        train['covariates'].columns.to_list() + ['hamster']
    ))

    try:
        dut.start()
        dut.fit(train)

        advance([dut])

        sock.sendto(match_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])
        assert_no_exceptions([dut])

        sock.sendto(mismatch_message, (LOOPBACK, COLS_RECEIVER_PORT))

        advance([dut])
        got = dut.poll_exceptions()
        while got is not None:
            assert isinstance(got[0], NotImplementedError), 'Should have raised NotImplementedError'
            assert 'mismatch' in str(got[1]), 'Should have mentioned mismatch'
            assert 'hamster' in str(got[1]), 'Should have mentioned hamster'
            got = dut.poll_exceptions()

        # make sure there are no other exceptions we don't expect
        assert_no_exceptions([dut])

    finally:
        dut.stop()
        sock.close()


NON_DISTRIBUTED_PORT = base_port.next()


def test_non_distributed(assert_no_exceptions: Clearer) -> None:
    '''Compare mean log PDF of our DT to sklearn's DT on one node.

    See: https://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    '''
    distributed_config = DistributedConfig(
        clause=build_clause(adjacency={'2': []},
                            nodes_and_endpoints=[('2', (LOOPBACK, NON_DISTRIBUTED_PORT))],
                            my_id=2))

    sklearn_alg = AlgorithmCatalogAuto().lookup_by_name('sklearn.tree.DecisionTreeClassifier')
    assert isinstance(sklearn_alg, SklearnAlgorithm), 'Should be a SklearnAlgorithm.'

    train, test, ground_truth, _ = load_classification_dataset()
    sklearn_train = deepcopy(train)

    alg = AutonDecisionTreeModel()

    dut = alg.instantiate(
        distributed=distributed_config,
        **DT_PARAMS)
    assert isinstance(dut, AutonDecisionTreeInstance), 'Should be an AutonDecisionTreeInstance.'

    sklearn_dut = sklearn_alg.instantiate(random_state=1701)

    acc = MetricCatalogAuto().lookup_by_name('accuracy_score')

    try:
        dut.start()

        sklearn_dut.fit(sklearn_train)
        dut.fit(train)

        got = dut.predict(test)
        sklearn_got = sklearn_dut.predict(test)

        got_acc = acc.calculate(pred=got, ground_truth=ground_truth)
        sklearn_acc = acc.calculate(pred=sklearn_got, ground_truth=ground_truth)
        assert got_acc == pytest.approx(sklearn_acc, rel=0.1), (
            'Should be close to sklearn accuracy.')

        assert_no_exceptions([dut])
    finally:
        dut.stop()


DISTRIBUTED_RECEIVER_PORT = base_port.next()
DISTRIBUTED_SENDER_PORT = base_port.next()


def test_distributed(assert_no_exceptions: Clearer) -> None:
    '''A node gets a minimal dataset to train on, test that it learns from an other.'''
    # pylint: disable=too-many-locals
    clause1 = build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[
            ('1', (LOOPBACK, DISTRIBUTED_RECEIVER_PORT)),
            ('2', (LOOPBACK, DISTRIBUTED_SENDER_PORT)),
        ],
        my_id=1
    )
    clause2 = deepcopy(clause1)
    clause2['my_id'] = 2

    train, test, ground_truth, reduced_train = load_classification_dataset()

    alg = AutonDecisionTreeModel()
    dut1 = alg.instantiate(
        distributed=DistributedConfig(clause1),
        synchronous=True,
        **DT_PARAMS)
    assert isinstance(dut1, AutonDecisionTreeInstance), 'Should be an AutonDecisionTreeInstance'

    acc = MetricCatalogAuto().lookup_by_name('accuracy_score')

    dut2 = None
    try:
        # fit dut 1 on reduced train set
        dut1.start()
        dut1.fit(reduced_train)

        advance([dut1])

        # check that dut1's training is bad
        got1 = dut1.predict(test)
        got1_acc = acc.calculate(pred=got1, ground_truth=ground_truth)
        assert got1_acc < 0.6, 'Should have a bad accuracy.'

        # fit dut2 on full train set
        dut2 = alg.instantiate(
            distributed=DistributedConfig(clause2),
            synchronous=True,
            **DT_PARAMS
        )
        assert isinstance(dut2, AutonDecisionTreeInstance), 'Should be an AutonDecisionTreeInstance'
        dut2.start()
        dut2.fit(train)
        for _ in range(5):
            advance([dut2, dut1])

        got1_improved = dut1.predict(test)
        got2 = dut2.predict(test)
        got1_improved_acc = acc.calculate(pred=got1_improved, ground_truth=ground_truth)
        got2_acc = acc.calculate(pred=got2, ground_truth=ground_truth)

        assert got1_improved_acc > got1_acc, 'Should have improved accuracy.'
        assert got1_improved_acc == pytest.approx(got2_acc, rel=1e-5), (
            'Should have similar accuracy.')

        assert_no_exceptions([dut1, dut2])

    finally:
        dut1.stop()
        if dut2 is not None:
            dut2.stop()


TWO_HALF_PORT_1 = base_port.next()
TWO_HALF_PORT_2 = base_port.next()


def test_two_half_hard_mode(assert_no_exceptions: Clearer) -> None:
    '''Two nodes with half the training behaves as (reasonably) expected.

    Hard mode means that we have completely dividided the dataset along
    a random hyperplane, so that the two halves are completely different.
    '''
    distributed_clause = build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[
            ('1', (LOOPBACK, TWO_HALF_PORT_1)),
            ('2', (LOOPBACK, TWO_HALF_PORT_2)),
        ],
        my_id=0,
        data_split={'num_nodes': 2}
    )
    train, _, _, _ = load_classification_dataset()

    # Split the training data such that each node sees a differently distributed half of the
    # input space.
    # Kyle's method to choose a random hyperplane and use it to divide data:
    # 1. Sample a random d-dimensional vector from a standard gaussian distribution.
    # 2. Project all of our data onto that vector - dot product of that vector with data matrix
    # gives you 1 number per row in the dataset
    # 3. sort based on that number and divide data in half
    input_df: pd.DataFrame = train['covariates']
    input_df['target'] = train['target']['target']
    threadsafe_random_state = np.random.RandomState(seed=Defaults.SEED)  # pylint: disable=no-member,line-too-long
    random_hyperplane = threadsafe_random_state.normal(size=input_df.shape[1])
    projection = np.dot(random_hyperplane, input_df.to_numpy().transpose())
    input_df['projection'] = projection
    input_df.sort_values(by='projection')
    input_df.drop(columns=['projection'], inplace=True)

    half_n = int(input_df.shape[0] / 2)

    train_first = Dataset(
        metadata=train.metadata,
        dataframe=input_df.iloc[:half_n]
    )
    train_second = Dataset(
        metadata=train.metadata,
        dataframe=input_df.iloc[half_n:]
    )
    distributed_receiver_clause = deepcopy(distributed_clause)
    distributed_receiver_clause['my_id'] = 1
    distributed_sender_clause = deepcopy(distributed_clause)
    distributed_sender_clause['my_id'] = 2

    params = deepcopy(DT_PARAMS)
    params['ccp_alpha'] = 1e-3

    alg = AutonDecisionTreeModel()
    dut_first = alg.instantiate(
        distributed=DistributedConfig(distributed_receiver_clause),
        synchronous=True,
        **params)
    assert isinstance(dut_first, AutonDecisionTreeInstance), (
        'dut_first should be an AutonDecisionTreeInstance')

    dut_second = alg.instantiate(
        distributed=DistributedConfig(distributed_sender_clause),
        synchronous=True,
        **params)
    assert isinstance(dut_second, AutonDecisionTreeInstance), (
        'dut_second should be an AutonDecisionTreeInstance'
    )

    try:
        dut_first.start()
        dut_second.start()

        dut_first.fit(train_first)
        dut_second.fit(train_second)

        for _ in range(20):
            advance([dut_first, dut_second])

        # Confirm that the two nodes have similar training and thus shared information.
        got = np.sqrt(dut_second.norm2_diff(dut_first))

        assert got == pytest.approx(0, abs=0.1), 'Should have similar training.'
        assert_no_exceptions([dut_first, dut_second])

    finally:
        dut_first.stop()
        dut_second.stop()


def test_two_same(assert_no_exceptions: Clearer) -> None:
    '''Two nodes with the same training don't change their coefficients.'''
    distributed_clause = build_clause(
        adjacency={'1': [2], '2': [1]},
        nodes_and_endpoints=[],
        my_id=0)

    distributed_clause['communicator'] = {
        'name': 'memory',
        'memory': {
            'domain': 'test_two_same',
        }
    }

    train, _, _, _ = load_classification_dataset()

    clause1 = deepcopy(distributed_clause)
    clause1['my_id'] = 1
    clause2 = deepcopy(distributed_clause)
    clause2['my_id'] = 2

    alg = AutonDecisionTreeModel()
    dut1 = alg.instantiate(
        distributed=DistributedConfig(clause1),
        synchronous=True,
        **DT_PARAMS)
    assert isinstance(dut1, AutonDecisionTreeInstance), (
        'dut1 should be an AutonDecisionTreeInstance')

    dut2 = alg.instantiate(
        distributed=DistributedConfig(clause2),
        synchronous=True,
        **DT_PARAMS)
    assert isinstance(dut2, AutonDecisionTreeInstance), (
        'dut2 should be an AutonDecisionTreeInstance')

    try:
        dut1.start()
        dut2.start()

        dut1.fit(train)
        dut2.fit(train)

        # Wait for the message to arrive and get processed.
        advance([dut1, dut2])

        # Confirm that neighbor optimization doesn't mess up result.
        got = np.sqrt(dut1.norm2_diff(dut2))

        assert got == pytest.approx(0, abs=0.1), 'Should have similar training.'
        assert_no_exceptions([dut2, dut1])

    finally:
        dut1.stop()
        dut2.stop()


TEST_N_SAME_N = 5
TEST_N_SAME_PORTS = [base_port.next() for _ in range(TEST_N_SAME_N)]
TEST_N_SAME_REFERENCE_PORT = base_port.next()


def test_n_same(wait_til_all_fit: Waiter,
                assert_no_exceptions: Clearer) -> None:
    '''N nodes fully connected with the same training match the non-distributed result.'''
    n = TEST_N_SAME_N

    adjacency = {f'{k}': [j for j in range(1, n + 1) if j != k] for k in range(1, n + 1)}
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_N_SAME_PORTS[k - 1])) for k in range(1, n + 1)]
    distributed_clause = build_clause(
        adjacency=adjacency, nodes_and_endpoints=nodes_and_endpoints, my_id=0
    )

    distributed_clause['communicator'] = {
        'name': 'memory',
        'memory': {
            'domain': 'test_n_same',
        }
    }

    train, _, _, _ = load_classification_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonDecisionTreeModel()

    duts: List[AutonDecisionTreeInstance] = [alg.instantiate(
        distributed=DistributedConfig(dc),
        **DT_PARAMS) for dc in distributed_clauses]
    try:
        for dut in duts:
            dut.start()

        for dut in duts:
            dut.fit(train)

        # Wait for the message to arrive and get processed.
        wait_til_all_fit(duts, convergence_check=True)

        # Confirm that neighbor optimization doesn't mess up result.

        dut0_state = duts[0].my_state_copy
        assert dut0_state is not None, 'dut0_state should not be None.'

        got = []
        for duti in duts[1:]:
            duti_state = duti.my_state_copy
            assert duti_state is not None, f'dut {dut._my_id} should not be None.'  # pylint: disable=protected-access
            got.append(np.sqrt(dut0_state.norm2_diff(duti_state)))
        assert all(got_diff < 1.0 for got_diff in got), 'Should have similar training.'
        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


TEST_TELEPHONE_N = 5
TEST_TELEPHONE_PORTS = [base_port.next() for _ in range(TEST_TELEPHONE_N)]
TEST_TELEPHONE_REFERENCE_PORT = base_port.next()


def test_telephone(wait_til_all_fit: Waiter,
                   assert_no_exceptions: Clearer) -> None:
    '''Whisper down the lane test. Nodes connected in a line.
    The only node that has data is one at the end '''
    n = TEST_TELEPHONE_N

    adjacency = {f'{k}': [k - 1, k + 1] for k in range(2, n)}
    adjacency['1'] = [2]
    adjacency[f'{n}'] = [n - 1]
    nodes_and_endpoints = [(f'{k}', (LOOPBACK, TEST_TELEPHONE_PORTS[k - 1]))
                           for k in range(1, n + 1)]
    distributed_clause = build_clause(
        adjacency=adjacency, nodes_and_endpoints=nodes_and_endpoints, my_id=0
    )

    distributed_clause['fit_eps'] = 0.001

    train, _, _, _ = load_classification_dataset()

    distributed_clauses = [dict(distributed_clause, my_id=k) for k in range(1, n + 1)]

    alg = AutonDecisionTreeModel()

    params = [dict(DT_PARAMS) for k in range(n)]

    duts: List[AutonDecisionTreeInstance] = [alg.instantiate(
        distributed=DistributedConfig(dc),
        **ps) for (dc, ps) in zip(distributed_clauses, params)]
    try:
        for dut in duts:
            dut.start()

        duts[0].fit(train)

        # Wait for the message to arrive and get processed.
        assert wait_til_all_fit(duts, convergence_check=True, max_time=60.0), (
            'Timeout while waiting for convergence'
        )

        # Confirm that neighbor optimization doesn't mess up result.
        dut0 = duts[0]
        got: List[np.float64] = []
        for duti in duts[1:]:
            got.append(np.sqrt(dut0.norm2_diff(duti)))

        want = [np.float64(0.0)] * len(got)

        np.testing.assert_allclose(actual=got, desired=want, atol=1.5)
        assert_no_exceptions(duts)

    finally:
        for dut in duts:
            dut.stop()


@pytest.mark.parametrize("target1,target2", [
    (['a', 'b', 'b'], ['a', 'b', 'c', 'c']),
    (['a', 'b', 'b'], ['b', 'c', 'd', 'c']),
    (['a', 'b', 'b', 'a'], ['d', 'c', 'c']),
    (['a'], ['a', 'b', 'b']),
])
def test_multiclass_mismatch(assert_no_exceptions: Clearer,
                             target1: List[str],
                             target2: List[str]) -> None:
    dist_clause_1 = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1]
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, base_port.next())),
                    ('2', (LOOPBACK, base_port.next())),
                ]
            },
        },
        'my_id': 1,
        'fit_eps': 0.1
    }

    dist_clause_2 = deepcopy(dist_clause_1)
    dist_clause_2['my_id'] = 2

    mismatch_params = deepcopy(DT_PARAMS)
    # With a Lambda of 1.0, each model oscillates between two states.
    mismatch_params['Lambda'] = 10

    alg = AutonDecisionTreeModel()
    dut1 = alg.instantiate(
        distributed=DistributedConfig(dist_clause_1),
        synchronous=True,
        **mismatch_params)
    assert isinstance(dut1, AutonDecisionTreeInstance), (
        'dut1 should be an AutonDecisionTreeInstance')

    dut2 = alg.instantiate(
        distributed=DistributedConfig(dist_clause_2),
        synchronous=True,
        **mismatch_params)
    assert isinstance(dut2, AutonDecisionTreeInstance), (
        'dut2 should be an AutonDecisionTreeInstance')

    state = np.random.default_rng(seed=1701)

    train_df_1 = pd.DataFrame({
        'cov1': state.random(size=len(target1)),
        'cov2': state.random(size=len(target1)),
        'target': target1
    })

    train_df_2 = pd.DataFrame({
        'cov1': state.random(size=len(target2)),
        'cov2': state.random(size=len(target2)),
        'target': target2
    })

    meta = Metadata(roles={
        RoleName.TARGET: [Column('target')]
    })

    ds_1 = Dataset(dataframe=train_df_1, metadata=meta)
    ds_2 = Dataset(dataframe=train_df_2, metadata=meta)

    try:
        dut1.start()
        # We want to make sure that dut2 doesn't get the initial messages from dut1.
        dut1.fit(ds_1)

        advance([dut1])

        dut2.start()
        dut2.fit(ds_2)

        advance([dut1, dut2])

        assert dut1.my_state is not None, 'dut1.my_state should not be None.'
        assert dut1.my_state.class_list == sorted(set().union(target1, target2)), (
            f'dut1.my_state.class_list: {dut1.my_state.class_list}, '
            f'should be the union of target1: {target1}, and target2: {target2}.')

        assert dut2.my_state is not None, 'dut2.my_state should not be None.'
        assert dut2.my_state.distance(dut1.my_state) == dut1.my_state.distance(dut2.my_state), (
            'distance calculate is not symmetric'
        )
        # It seems that the above assert caught the bug that caused the error below.
        # dut1.fit(ds_2)  # This refit causes a test to fail, for some reason.
        # assert wait_til_all_fit([dut1, dut2], convergence_check=True)

        assert_no_exceptions([dut1, dut2])
    finally:
        dut1.stop()
        dut2.stop()


# TODO(Merritt/Piggy): add this test to every distributed algorithm
def test_distance_axioms(wait_til_all_fit: Waiter,
                         assert_no_exceptions: Clearer) -> None:
    '''Commutative: dist(a, b) = dist(b, a)
    Non-negative: dist(a, b) >= 0
    dist(a, a) == 0
    dist(a, b) + dist (b, c) > dist(a, c)
    '''
    # pylint: disable=too-many-statements
    dist_clause_1 = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [],
                    '2': [],
                    '3': [],
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, base_port.next())),
                    ('2', (LOOPBACK, base_port.next())),
                    ('3', (LOOPBACK, base_port.next())),
                ]
            },
        },
        'my_id': 1,
        'fit_eps': 1e-4
    }

    dist_clause_2 = deepcopy(dist_clause_1)
    dist_clause_2['my_id'] = 2

    dist_clause_3 = deepcopy(dist_clause_1)
    dist_clause_3['my_id'] = 3

    alg = AutonDecisionTreeModel()
    dut1 = alg.instantiate(
        distributed=DistributedConfig(dist_clause_1),
        **DT_PARAMS)
    assert isinstance(dut1, AutonDecisionTreeInstance), (
        'dut1 should be an AutonDecisionTreeInstance')

    dut2 = alg.instantiate(
        distributed=DistributedConfig(dist_clause_2),
        **DT_PARAMS)
    assert isinstance(dut2, AutonDecisionTreeInstance), (
        'dut2 should be an AutonDecisionTreeInstance')

    dut3 = alg.instantiate(
        distributed=DistributedConfig(dist_clause_3),
        **DT_PARAMS)
    assert isinstance(dut3, AutonDecisionTreeInstance), (
        'dut3 should be an AutonDecisionTreeInstance'
    )

    state = np.random.default_rng(seed=1701)

    target1 = ['a', 'b', 'c', 'a']
    train_df_1 = pd.DataFrame({
        'cov1': state.random(size=len(target1)),
        'cov2': state.random(size=len(target1)),
        'target': target1
    })

    target2 = ['b', 'a', 'b', 'c', 'c']
    train_df_2 = pd.DataFrame({
        'cov1': state.random(size=len(target2)),
        'cov2': state.random(size=len(target2)),
        'target': target2
    })

    target3 = ['a', 'b', 'c', 'c']
    train_df_3 = pd.DataFrame({
        'cov1': state.random(size=len(target3)),
        'cov2': state.random(size=len(target3)),
        'target': target3
    })

    meta = Metadata(roles={
        RoleName.TARGET: [Column('target')]
    })

    ds_1 = Dataset(dataframe=train_df_1, metadata=meta)
    ds_2 = Dataset(dataframe=train_df_2, metadata=meta)
    ds_3 = Dataset(dataframe=train_df_3, metadata=meta)

    try:
        dut1.start()
        dut2.start()
        dut3.start()

        dut1.fit(ds_1)
        dut2.fit(ds_2)
        dut3.fit(ds_3)

        wait_til_all_fit([dut1, dut2, dut3], convergence_check=False)

        assert dut1.my_state is not None, 'dut1.my_state should not be None.'
        assert dut2.my_state is not None, 'dut2.my_state should not be None.'
        assert dut3.my_state is not None, 'dut3.my_state should not be None.'
        distance_1_1 = dut1.my_state.distance(dut1.my_state)
        distance_1_2 = dut1.my_state.distance(dut2.my_state)
        distance_1_3 = dut1.my_state.distance(dut3.my_state)
        distance_2_1 = dut2.my_state.distance(dut1.my_state)
        distance_2_2 = dut2.my_state.distance(dut2.my_state)
        distance_2_3 = dut2.my_state.distance(dut3.my_state)
        distance_3_1 = dut3.my_state.distance(dut1.my_state)
        distance_3_2 = dut3.my_state.distance(dut2.my_state)
        distance_3_3 = dut3.my_state.distance(dut3.my_state)

        # Identity
        assert distance_1_1 == 0, f'ident distance_1_1: {distance_1_1}'
        assert distance_2_2 == 0, f'ident distance_2_2: {distance_2_2}'
        assert distance_3_3 == 0, f'ident distance_3_3: {distance_3_3}'

        # Commutativity
        assert distance_1_2 == distance_2_1, (
            f'commut distance_1_2: {distance_1_2}, distance_2_1: {distance_2_1}')
        assert distance_1_3 == distance_3_1, (
            f'commut distance_1_3: {distance_1_3}, distance_3_1: {distance_3_1}')
        assert distance_2_3 == distance_3_2, (
            f'commut distance_2_3: {distance_2_3}, distance_3_2: {distance_3_2}')

        # Non-negativity
        assert distance_1_2 > 0, f'non-neg distance_1_2: {distance_1_2}'
        assert distance_1_3 > 0, f'non-neg distance_1_3: {distance_1_3}'
        assert distance_2_3 > 0, f'non-neg distance_2_3: {distance_2_3}'

        # Triangle inequality
        assert distance_1_2 + distance_2_3 >= distance_1_3, (
            f'triangle distance_1_2: {distance_1_2}, distance_2_3: {distance_2_3}, '
            f'distance_1_3: {distance_1_3}')
        assert distance_1_3 + distance_3_2 >= distance_1_2, (
            f'triangle distance_1_3: {distance_1_3}, distance_3_2: {distance_3_2}, '
            f'distance_1_2: {distance_1_2}')
        assert distance_2_3 + distance_3_1 >= distance_2_1, (
            f'triangle distance_2_3: {distance_2_3}, distance_3_1: {distance_3_1}, '
            f'distance_2_1: {distance_2_1}')
        assert distance_2_1 + distance_1_3 >= distance_2_3, (
            f'triangle distance_2_1: {distance_2_1}, distance_1_3: {distance_1_3}, '
            f'distance_2_3: {distance_2_3}')
        assert distance_3_2 + distance_2_1 >= distance_3_1, (
            f'triangle distance_3_2: {distance_3_2}, distance_2_1: {distance_2_1}, '
            f'distance_3_1: {distance_3_1}')
        assert distance_3_1 + distance_1_2 >= distance_3_2, (
            f'triangle distance_3_1: {distance_3_1}, distance_1_2: {distance_1_2}, '
            f'distance_3_2: {distance_3_2}')

        assert_no_exceptions([dut1, dut2, dut3])
    finally:
        dut1.stop()
        dut2.stop()
        dut3.stop()


def load_multiclass_dataset() -> Tuple[Dataset, Dataset, Dataset, Dataset]:
    # Load the iris dataset
    bc_x_full, bc_y_series = datasets.load_iris(return_X_y=True, as_frame=True)
    assert isinstance(bc_x_full, pd.DataFrame)
    assert isinstance(bc_y_series, pd.Series)
    bc_y = pd.DataFrame({'target': bc_y_series})

    # restrict number of attributes for wider variability of results
    bc_x = bc_x_full.iloc[:, :3]

    # shuffle our data
    bc_x = bc_x.sample(frac=1, random_state=1701)
    bc_y = bc_y.sample(frac=1, random_state=1701)

    test_size = 50
    # Split the data into training/testing sets
    bc_x_train = bc_x[:-test_size]
    bc_x_test = bc_x[-test_size:]

    # Split the targets into training/testing sets
    bc_y_train = bc_y[:-test_size]
    bc_y_test = bc_y[-test_size:]

    metadata = Metadata(
        roles={RoleName.TARGET: [Column('target')]},
        task=TaskType.BINARY_CLASSIFICATION
    )

    dataset_train = Dataset(
        metadata=metadata,
        covariates=bc_x_train,
        target=bc_y_train
    )

    dataset_one_point_train = Dataset(
        metadata=metadata,
        covariates=bc_x_train[:1],
        target=bc_y_train[:1]
    )

    dataset_test = Dataset(
        metadata=metadata,
        covariates=bc_x_test
    )
    ground_truth = Dataset(
        metadata=metadata,
        ground_truth=bc_y_test
    )

    assert len(np.unique(bc_y_train)) == 3
    return (dataset_train, dataset_one_point_train, dataset_test, ground_truth)


def test_real_multiclass_data(wait_til_all_fit: Waiter,
                              assert_no_exceptions: Clearer) -> None:
    dist_clause_1 = {
        'discoverer': {
            'name': 'static',
            'static': {
                'adjacency': {
                    '1': [2],
                    '2': [1]
                }
            }
        },
        'communicator': {
            'name': 'sockets',
            'sockets': {
                'nodes_and_endpoints': [
                    ('1', (LOOPBACK, base_port.next())),
                    ('2', (LOOPBACK, base_port.next())),
                ]
            },
        },
        'my_id': 1,
        'fit_eps': 0.1
    }

    dist_clause_2 = deepcopy(dist_clause_1)
    dist_clause_2['my_id'] = 2

    alg = AutonDecisionTreeModel()
    dut1 = alg.instantiate(
        distributed=DistributedConfig(dist_clause_1),
        **DT_PARAMS)
    assert isinstance(dut1, AutonDecisionTreeInstance), (
        'dut1 should be an AutonDecisionTreeInstance')

    dut2 = alg.instantiate(
        distributed=DistributedConfig(dist_clause_2),
        **DT_PARAMS)
    assert isinstance(dut2, AutonDecisionTreeInstance), (
        'dut2 should be an AutonDecisionTreeInstance')

    train, one_pt_train, test, ground_truth = load_multiclass_dataset()

    acc = MetricCatalogAuto().lookup_by_name('accuracy_score')
    assert isinstance(acc, Metric)

    try:
        dut1.start()
        dut2.start()

        dut1.fit(one_pt_train)

        assert wait_til_all_fit([dut1, dut2], convergence_check=False), (
            'Timeout while waiting for first fit')

        dut2.fit(train)

        assert wait_til_all_fit([dut1, dut2], convergence_check=True), (
            'Timeout while waiting for second fit to converge')

        pred1 = dut1.predict(test)
        assert pred1 is not None, 'pred1 should not be None.'
        pred2 = dut2.predict(test)
        assert pred2 is not None, 'pred2 should not be None.'

        acc1 = acc.calculate(pred=pred1, ground_truth=ground_truth)
        acc2 = acc.calculate(pred=pred2, ground_truth=ground_truth)

        num_classes = 3.0
        random_guessing_acc = 1.0 / num_classes
        assert acc1 > random_guessing_acc, (
            f'Accuracy of first model ({acc1}) '
            f'should be better than random guessing ({random_guessing_acc})'
        )
        assert acc2 > random_guessing_acc, (
            f'Accuracy of second model ({acc2}) '
            f'should be better than random guessing ({random_guessing_acc})'
        )

        assert_no_exceptions([dut1, dut2])
    finally:
        dut1.stop()
        dut2.stop()



'''Tests for dict.py.'''

# Copyright (c) 2025 Carnegie Mellon University
# This code is subject to the license terms contained in the LICENSE file.

import numpy as np
import pandas as pd

from .dict import DictTable
from .impl.table_auto import TableCatalogAuto
from .impl.table import TableFactory


def test_dataframe_table() -> None:
    '''Test the DataFrameTable.'''
    _ = TableCatalogAuto()

    my_dict = {'a': [1, 2, 3], 'b': [4, 5, 6]}
    my_table = TableFactory(my_dict)  # type: ignore[abstract] # pylint: disable=abstract-class-instantiated

    # assert isinstance(my_table, Table)
    assert isinstance(my_table, DictTable)

    assert list(my_table['a']) == [1, 2, 3]
    assert list(my_table['b']) == [4, 5, 6]

    # TODO(Merritt/Piggy): Make this test work. Possibly return a pd.Series instead of a list.
    # my_table['c'] = my_table['a'] + my_table['b']
    # assert list(my_table['c']) == [5, 7, 9]

    assert my_table.as_(dict) == my_dict

    assert my_table.as_(pd.DataFrame).iloc[1, 1] == 5

    my_ndarray = my_table.as_(np.ndarray)
    assert isinstance(my_ndarray, np.ndarray)
    assert my_ndarray.shape == (3, 2)
